import pandas as pd
import numpy as np
import biogeme.database as db
import biogeme.biogeme as bio
#import biogeme.models as models
import biogeme.loglikelihood as ll
from biogeme.expressions import Beta, DefineVariable, log, Integrate, Elem, bioNormalCdf

pandas = pd.read_csv("optima.dat",sep='\t')
database = db.Database("optima",pandas)

globals().update(database.variables)

exclude = (Choice == -1.0)
database.remove(exclude)



### Variables

ScaledIncome = DefineVariable('ScaledIncome',\
                              CalculatedIncome / 1000,database)

formulaIncome = models.piecewiseFormula(ScaledIncome,[None,4,6,8,10,None],[0.0902761,-0.221283,0.259466,-0.523049,0.084351])

age_65_more = DefineVariable('age_65_more',age >= Numeric(65),database)
moreThanOneCar = DefineVariable('moreThanOneCar',NbCar > 1,database)
moreThanOneBike = DefineVariable('moreThanOneBike',NbBicy > 1,database)
individualHouse = DefineVariable('individualHouse',\
                                 HouseType == 1,database)
male = DefineVariable('male',Gender == 1,database)
haveChildren = DefineVariable('haveChildren',\
                              ((FamilSitu == 3)+(FamilSitu == 4)) > 0,database)
haveGA = DefineVariable('haveGA',GenAbST == 1,database)
highEducation = DefineVariable('highEducation', Education >= 6,database)

### Coefficients
coef_intercept = Beta('coef_intercept',0.398165,None,None,0 )
coef_age_65_more = Beta('coef_age_65_more',0.0716533,None,None,0 )
coef_haveGA = Beta('coef_haveGA',-0.578005,None,None,0 )
coef_moreThanOneCar = \
 Beta('coef_moreThanOneCar',0.53301,None,None,0 )
coef_moreThanOneBike = \
 Beta('coef_moreThanOneBike',-0.277122,None,None,0 )
coef_individualHouse = \
 Beta('coef_individualHouse',-0.0885649,None,None,0 )
coef_male = Beta('coef_male',0.0663476,None,None,0 )
coef_haveChildren = Beta('coef_haveChildren',-0.0376042,None,None,0 )
coef_highEducation = Beta('coef_highEducation',-0.246687,None,None,0 )

### Latent variable: structural equation

# Note that the expression must be on a single line. In order to 
# write it across several lines, each line must terminate with 
# the \ symbol

CARLOVERS = \
coef_intercept +\
coef_age_65_more * age_65_more +\
formulaIncome+\
coef_moreThanOneCar * moreThanOneCar +\
coef_moreThanOneBike * moreThanOneBike +\
coef_individualHouse * individualHouse +\
coef_male * male +\
coef_haveChildren * haveChildren +\
coef_haveGA * haveGA +\
coef_highEducation * highEducation


### Measurement equations

INTER_Envir01 = Beta('INTER_Envir01',0,None,None,1)
INTER_Envir02 = Beta('INTER_Envir02',0.348654,None,None,0 )
INTER_Envir03 = Beta('INTER_Envir03',-0.309023,None,None,0 )
INTER_Mobil11 = Beta('INTER_Mobil11',0.337726,None,None,0 )
INTER_Mobil14 = Beta('INTER_Mobil14',-0.130563,None,None,0 )
INTER_Mobil16 = Beta('INTER_Mobil16',0.128293,None,None,0 )
INTER_Mobil17 = Beta('INTER_Mobil17',0.145876,None,None,0 )

B_Envir01_F1 = Beta('B_Envir01_F1',-1,None,None,1)
B_Envir02_F1 = Beta('B_Envir02_F1',-0.431461,None,None,0 )
B_Envir03_F1 = Beta('B_Envir03_F1',0.565903,None,None,0 )
B_Mobil11_F1 = Beta('B_Mobil11_F1',0.483958,None,None,0 )
B_Mobil14_F1 = Beta('B_Mobil14_F1',0.58221,None,None,0 )
B_Mobil16_F1 = Beta('B_Mobil16_F1',0.463139,None,None,0 )
B_Mobil17_F1 = Beta('B_Mobil17_F1',0.368257,None,None,0 )



MODEL_Envir01 = INTER_Envir01 + B_Envir01_F1 * CARLOVERS
MODEL_Envir02 = INTER_Envir02 + B_Envir02_F1 * CARLOVERS
MODEL_Envir03 = INTER_Envir03 + B_Envir03_F1 * CARLOVERS
MODEL_Mobil11 = INTER_Mobil11 + B_Mobil11_F1 * CARLOVERS
MODEL_Mobil14 = INTER_Mobil14 + B_Mobil14_F1 * CARLOVERS
MODEL_Mobil16 = INTER_Mobil16 + B_Mobil16_F1 * CARLOVERS
MODEL_Mobil17 = INTER_Mobil17 + B_Mobil17_F1 * CARLOVERS

SIGMA_STAR_Envir01 = Beta('SIGMA_STAR_Envir01',1,-10000,10000,1)
SIGMA_STAR_Envir02 = Beta('SIGMA_STAR_Envir02',0.767063,-10000,10000,0 )
SIGMA_STAR_Envir03 = Beta('SIGMA_STAR_Envir03',0.717835,-10000,10000,0 )
SIGMA_STAR_Mobil11 = Beta('SIGMA_STAR_Mobil11',0.783358,-10000,10000,0 )
SIGMA_STAR_Mobil14 = Beta('SIGMA_STAR_Mobil14',0.688264,-10000,10000,0 )
SIGMA_STAR_Mobil16 = Beta('SIGMA_STAR_Mobil16',0.754419,-10000,10000,0 )
SIGMA_STAR_Mobil17 = Beta('SIGMA_STAR_Mobil17',0.760104,-10000,10000,0 )

delta_1 = Beta('delta_1',0.251983,0,10,0 )
delta_2 = Beta('delta_2',0.759208,0,10,0 )
tau_1 = -delta_1 - delta_2
tau_2 = -delta_1 
tau_3 = delta_1
tau_4 = delta_1 + delta_2

Envir01_tau_1 = (tau_1-MODEL_Envir01) / SIGMA_STAR_Envir01
Envir01_tau_2 = (tau_2-MODEL_Envir01) / SIGMA_STAR_Envir01
Envir01_tau_3 = (tau_3-MODEL_Envir01) / SIGMA_STAR_Envir01
Envir01_tau_4 = (tau_4-MODEL_Envir01) / SIGMA_STAR_Envir01
IndEnvir01 = {
    1: bioNormalCdf(Envir01_tau_1),
    2: bioNormalCdf(Envir01_tau_2)-bioNormalCdf(Envir01_tau_1),
    3: bioNormalCdf(Envir01_tau_3)-bioNormalCdf(Envir01_tau_2),
    4: bioNormalCdf(Envir01_tau_4)-bioNormalCdf(Envir01_tau_3),
    5: 1-bioNormalCdf(Envir01_tau_4),
    6: 1.0,
    -1: 1.0,
    -2: 1.0
}

P_Envir01 = Elem(IndEnvir01, Envir01)


Envir02_tau_1 = (tau_1-MODEL_Envir02) / SIGMA_STAR_Envir02
Envir02_tau_2 = (tau_2-MODEL_Envir02) / SIGMA_STAR_Envir02
Envir02_tau_3 = (tau_3-MODEL_Envir02) / SIGMA_STAR_Envir02
Envir02_tau_4 = (tau_4-MODEL_Envir02) / SIGMA_STAR_Envir02
IndEnvir02 = {
    1: bioNormalCdf(Envir02_tau_1),
    2: bioNormalCdf(Envir02_tau_2)-bioNormalCdf(Envir02_tau_1),
    3: bioNormalCdf(Envir02_tau_3)-bioNormalCdf(Envir02_tau_2),
    4: bioNormalCdf(Envir02_tau_4)-bioNormalCdf(Envir02_tau_3),
    5: 1-bioNormalCdf(Envir02_tau_4),
    6: 1.0,
    -1: 1.0,
    -2: 1.0
}

P_Envir02 = Elem(IndEnvir02, Envir02)

Envir03_tau_1 = (tau_1-MODEL_Envir03) / SIGMA_STAR_Envir03
Envir03_tau_2 = (tau_2-MODEL_Envir03) / SIGMA_STAR_Envir03
Envir03_tau_3 = (tau_3-MODEL_Envir03) / SIGMA_STAR_Envir03
Envir03_tau_4 = (tau_4-MODEL_Envir03) / SIGMA_STAR_Envir03
IndEnvir03 = {
    1: bioNormalCdf(Envir03_tau_1),
    2: bioNormalCdf(Envir03_tau_2)-bioNormalCdf(Envir03_tau_1),
    3: bioNormalCdf(Envir03_tau_3)-bioNormalCdf(Envir03_tau_2),
    4: bioNormalCdf(Envir03_tau_4)-bioNormalCdf(Envir03_tau_3),
    5: 1-bioNormalCdf(Envir03_tau_4),
    6: 1.0,
    -1: 1.0,
    -2: 1.0
}

P_Envir03 = Elem(IndEnvir03, Envir03)

Mobil11_tau_1 = (tau_1-MODEL_Mobil11) / SIGMA_STAR_Mobil11
Mobil11_tau_2 = (tau_2-MODEL_Mobil11) / SIGMA_STAR_Mobil11
Mobil11_tau_3 = (tau_3-MODEL_Mobil11) / SIGMA_STAR_Mobil11
Mobil11_tau_4 = (tau_4-MODEL_Mobil11) / SIGMA_STAR_Mobil11
IndMobil11 = {
    1: bioNormalCdf(Mobil11_tau_1),
    2: bioNormalCdf(Mobil11_tau_2)-bioNormalCdf(Mobil11_tau_1),
    3: bioNormalCdf(Mobil11_tau_3)-bioNormalCdf(Mobil11_tau_2),
    4: bioNormalCdf(Mobil11_tau_4)-bioNormalCdf(Mobil11_tau_3),
    5: 1-bioNormalCdf(Mobil11_tau_4),
    6: 1.0,
    -1: 1.0,
    -2: 1.0
}

P_Mobil11 = Elem(IndMobil11, Mobil11)

Mobil14_tau_1 = (tau_1-MODEL_Mobil14) / SIGMA_STAR_Mobil14
Mobil14_tau_2 = (tau_2-MODEL_Mobil14) / SIGMA_STAR_Mobil14
Mobil14_tau_3 = (tau_3-MODEL_Mobil14) / SIGMA_STAR_Mobil14
Mobil14_tau_4 = (tau_4-MODEL_Mobil14) / SIGMA_STAR_Mobil14
IndMobil14 = {
    1: bioNormalCdf(Mobil14_tau_1),
    2: bioNormalCdf(Mobil14_tau_2)-bioNormalCdf(Mobil14_tau_1),
    3: bioNormalCdf(Mobil14_tau_3)-bioNormalCdf(Mobil14_tau_2),
    4: bioNormalCdf(Mobil14_tau_4)-bioNormalCdf(Mobil14_tau_3),
    5: 1-bioNormalCdf(Mobil14_tau_4),
    6: 1.0,
    -1: 1.0,
    -2: 1.0
}

P_Mobil14 = Elem(IndMobil14, Mobil14)

Mobil16_tau_1 = (tau_1-MODEL_Mobil16) / SIGMA_STAR_Mobil16
Mobil16_tau_2 = (tau_2-MODEL_Mobil16) / SIGMA_STAR_Mobil16
Mobil16_tau_3 = (tau_3-MODEL_Mobil16) / SIGMA_STAR_Mobil16
Mobil16_tau_4 = (tau_4-MODEL_Mobil16) / SIGMA_STAR_Mobil16
IndMobil16 = {
    1: bioNormalCdf(Mobil16_tau_1),
    2: bioNormalCdf(Mobil16_tau_2)-bioNormalCdf(Mobil16_tau_1),
    3: bioNormalCdf(Mobil16_tau_3)-bioNormalCdf(Mobil16_tau_2),
    4: bioNormalCdf(Mobil16_tau_4)-bioNormalCdf(Mobil16_tau_3),
    5: 1-bioNormalCdf(Mobil16_tau_4),
    6: 1.0,
    -1: 1.0,
    -2: 1.0
}

P_Mobil16 = Elem(IndMobil16, Mobil16)

Mobil17_tau_1 = (tau_1-MODEL_Mobil17) / SIGMA_STAR_Mobil17
Mobil17_tau_2 = (tau_2-MODEL_Mobil17) / SIGMA_STAR_Mobil17
Mobil17_tau_3 = (tau_3-MODEL_Mobil17) / SIGMA_STAR_Mobil17
Mobil17_tau_4 = (tau_4-MODEL_Mobil17) / SIGMA_STAR_Mobil17
IndMobil17 = {
    1: bioNormalCdf(Mobil17_tau_1),
    2: bioNormalCdf(Mobil17_tau_2)-bioNormalCdf(Mobil17_tau_1),
    3: bioNormalCdf(Mobil17_tau_3)-bioNormalCdf(Mobil17_tau_2),
    4: bioNormalCdf(Mobil17_tau_4)-bioNormalCdf(Mobil17_tau_3),
    5: 1-bioNormalCdf(Mobil17_tau_4),
    6: 1.0,
    -1: 1.0,
    -2: 1.0
}

P_Mobil17 = Elem(IndMobil17, Mobil17)


loglike = log(P_Envir01) + \
          log(P_Envir02) + \
          log(P_Envir03) + \
          log(P_Mobil11) + \
          log(P_Mobil14) + \
          log(P_Mobil16) + \
          log(P_Mobil17)


print(database.data.head())
biogeme  = bio.BIOGEME(database,loglike)
biogeme.modelName = "02oneLatentOrdered"
results = biogeme.estimate()
print("Estimated betas: {}".format(len(results.data.betaValues)))




"""Implementation of the main Biogeme class that combines the database and the model specification.

:author: Michel Bierlaire
:date: Tue Mar 26 16:45:15 2019

"""

import numpy as np
import pandas as pd
import pickle
from datetime import datetime
import multiprocessing as mp

import biogeme.database as db
import biogeme.cbiogeme as cb
import biogeme.expressions as eb
import biogeme.tools as tools
import biogeme.version as bv
import biogeme.results as res
import biogeme.exceptions as excep
import biogeme.filenames as bf
import biogeme.messaging as msg
import biogeme.optimization as opt

import biogeme.messaging as msg

logger = msg.bioMessage()

#import yep

class BIOGEME:
    """Main class that combines the database and the model specification.

    It works in two modes: estimation and simulation.

    """

    def __init__(self,
                 database,
                 formulas,
                 numberOfThreads=None,
                 numberOfDraws=1000,
                 seed=None,
                 skipAudit=False,
                 removeUnusedVariables=True,
                 missingData=99999):
        """Constructor

        :param database: choice data.
        :type database: biogeme.database

        :param formulas: expression or dictionary of expressions that
             define the model specification.  The concept is that each
             expression is applied to each entry of the database. The
             keys of the dictionary allow to provide a name to each
             formula.  In the estimation mode, two formulas are
             needed, with the keys 'loglike' and 'weight'. If only one
             formula is provided, it is associated with the label
             'loglike'. If no formula is labeled 'weight', the weight
             of each piece of data is supposed to be 1.0. In the
             simulation mode, the labels of each formula are used as
             labels of the resulting database.
        :type formulas: biogeme.expressions.Expression, or dict(biogeme.expressions.Expression)

        :param numberOfThreads: multi-threading can be used for
            estimation. This parameter defines the number of threads
            to be used. If the parameter is set to None, the number of
            available threads is calculated using
            cpu_count(). Ignored in simulation mode. Defaults: None.
        :type numberOfThreads:  int

        :param numberOfDraws: number of draws used for Monte-Carlo
            integration. Default: 1000.
        :type numberOfDraws: int

        :param seed: seed used for the pseudo-random number
            generation. It is useful only when each run should
            generate the exact same result. If None, a new seed is
            used at each run. Default: None.
        :type seed: int

        :param skipAudit: if True, does not check the validity of the
            formulas. It may save significant amount of time for large
            models and large data sets. Default: False.
        :type skipAudit: bool

        :param removeUnusedVariables: if True, all variables not used
           in the expression are removed from the database. Default:
           True.
        :type removeUnusedVariables: bool

        :param missingData: if one variable has this value, it is
           assumed that a data is missing and an exception will be
           triggered. Default: 99999.
        :type missingData: float

        """

        ## Logger that controls the output of messages to the screen and log file.
        self.logger = msg.bioMessage()
        if not skipAudit:
            database.data = database.data.replace({True:1, False:0})
            listOfErrors, listOfWarnings = database._audit()
            if listOfWarnings:
                self.logger.warning('\n'.join(listOfWarnings))
            if listOfErrors:
                self.logger.warning('\n'.join(listOfErrors))
                raise excep.biogemeError("\n".join(listOfErrors))

        ## Keyword used for the name of the loglikelihood formula. Default: 'loglike'
        self.loglikeName = 'loglike'
        ## Keyword used for the name of the weight formula. Default: 'weight'
        self.weightName = 'weight'
        ## Name of the model. Default: 'biogemeModelDefaultName'
        self.modelName = 'biogemeModelDefaultName'
        ## Comments to be added to the report
        self.comments = None
        ## monteCarlo is True if one of the expression involves a
        # Monte-Carlo integration.
        self.monteCarlo = False
        np.random.seed(seed)
        ## If True, the values are saved on a file each time the likelihood function is calculated
        self.saveIterations = False
        if not isinstance(formulas,dict):

            ## Object of type biogeme.expressions.Expression
            ## calculating the formula for the loglikelihood
            self.loglike = formulas

            ## Object of type biogeme.expressions.Expression
            ## calculating the weight of each observation in the
            ## sample
            self.weight = None

            ## Dictionary containing Biogeme formulas of type
            ## biogeme.expressions.Expression.
            # The keys are the names of the formulas.
            self.formulas = dict({self.loglikeName:formulas})
        else:
            self.loglike = formulas.get(self.loglikeName)
            self.weight = formulas.get(self.weightName)
            self.formulas = formulas

        ## biogeme.database object
        self.database = database

        ## Missing data
        self.missingData = missingData

        ## keep track of the sample of data used to calculate the
        ## stochastic gradient / hessian
        self.lastSample = None

        ## Init value of the likelihood function
        self.initLogLike = None

        if removeUnusedVariables:
            usedVariables = set()
            for k, f in self.formulas.items():
                usedVariables = usedVariables.union(f.setOfVariables())
            if self.database.isPanel():
                usedVariables.add(self.database.panelColumn)
            unusedVariables = set(self.database.data.columns) - usedVariables
            self.logger.general(f"Remove {len(unusedVariables)} unused variables from the database as only {len(usedVariables)} are used.")
            self.database.data = self.database.data.drop(columns=list(unusedVariables))


        if not skipAudit:
            self._audit()

        self.theC = cb.pyBiogeme()

        self._prepareDatabaseForFormula()
        self._prepareLiterals()
        ## Boolean variable, True if the HTML file with the results must be generated.
        self.generateHtml = True
        ## Boolean variable, True if the pickle file with the results must be generated.
        self.generatePickle = True

        ## Name of the column defining weights for batch sampling in
        ## stochastic optimization.
        self.columnForBatchSamplingWeights = None
        
        ## Number of threads used for parallel computing. Default: the number of CPU available.
        self.numberOfThreads = mp.cpu_count() if numberOfThreads is None else numberOfThreads
        start_time = datetime.now()
        self._generateDraws(numberOfDraws)
        if self.monteCarlo:
            self.theC.setDraws(self.database.theDraws)
        ## Time needed to generate the draws.
        self.drawsProcessingTime = datetime.now() - start_time
        if self.loglike is not None:

            ## Internal signature of the formula for the loglikelihood
            self.loglikeSignatures = self.loglike.getSignature()
            if self.weight is None:
                self.theC.setExpressions(self.loglikeSignatures,
                                         self.numberOfThreads)
            else:
                ## Internal signature of the formula for the weight
                self.weightSignatures = self.weight.getSignature()
                self.theC.setExpressions(self.loglikeSignatures,
                                         self.numberOfThreads,
                                         self.weightSignatures)


    def _audit(self):
        """Each expression provides an audit function, that verifies its
           validity. Each formula is audited, and the list of errors
           and warnings reported.

           :raise biogemeError: if the formula has issues, an error is
                                detected and an exception is raised.

        """

        listOfErrors = []
        listOfWarnings = []
        for k, v in self.formulas.items():
            err, war = v.audit(self.database)
            listOfErrors += err
            listOfWarnings += war
        if listOfWarnings:
            self.logger.warning('\n'.join(listOfWarnings))
        if listOfErrors:
            self.logger.warning('\n'.join(listOfErrors))
            raise excep.biogemeError("\n".join(listOfErrors))

    def _generateDraws(self, numberOfDraws):
        """If Monte-Carlo integration is involved in one of the formulas, this
           function instructs the database to generate the draws.

        Args:
            numberOfDraws: self explanatory (int)
        """


        ## Number of draws for Monte-Carlo integration.
        self.numberOfDraws = numberOfDraws
#        drawTypes = {}
#        for k,v in self.formulas.items() :
#            d = v.getDraws()
#            if d:
#                drawTypes = dict(drawTypes,**d)
        ## Draws
        self.monteCarlo = len(self.allDraws) > 0
        if self.monteCarlo:
            self.database.generateDraws(self.allDraws,
                                        self.drawNames,
                                        numberOfDraws)

            
    def _prepareDatabaseForFormula(self,sample=None):
        # Prepare the dataset. 
        if sample is None:
            if self.lastSample == 1.0:
                # We continue to use the full data set. Nothing to be done.
                return
            else:
                self.lastSample = 1.0
                self.database.useFullSample()
        else:
            # Check if the sample size is valid
            if sample <= 0 or sample > 1.0:
                raise ValueError(f'The value of the parameter sample must be strictly between 0.0 and 1.0, and not {sample}')

            if sample == 1.0:
                self.database.useFullSample()
            else:
                self.logger.detailed(f'Use {100*sample}% of the data.')
                if self.database.isPanel():
                    self.database.sampleIndividualMapWithoutReplacement(sample,self.columnForBatchSamplingWeights)
                else:
                    self.database.sampleWithoutReplacement(sample,self.columnForBatchSamplingWeights)
            self.lastSample = sample

        # Rebuild the map for panel data
        if self.database.isPanel():
            self.database._buildPanelMap()
            self.theC.setPanel(True)
            self.theC.setDataMap(self.database.individualMap)

        # Transfer the data to the C++ formula
        self.theC.setData(self.database.data)
        self.theC.setMissingData(self.missingData)

    def getBoundsOnBeta(self, betaName):
        """ Returns the bounds on the parameter as defined by the user.

        :param betaName: name of the parameter
        :type betaName: string
        :return: lower bound, upper bound
        :rtype: tuple
        :raises biogemeError: if the name of the parameter is not found.
        """

        if betaName not in self.freeBetaNames:
            raise excep.biogemeError(f"Unknown parameter {betaName}")
        index = self.freeBetaNames.index(betaName)
        return self.bounds[index]

    def _prepareLiterals(self):
        """ Extract from the formulas the literals (parameters,
        variables, random variables) and decide a numbering convention.
        """

        collectionOfFormulas = [f for k, f in self.formulas.items()]
        variableNames = list(self.database.data.columns.values)


        self.elementaryExpressionIndex, \
        self.allFreeBetas, \
        self.freeBetaNames, \
        self.allFixedBetas, \
        self.fixedBetaNames, \
        self.allRandomVariables, \
        self.randomVariableNames, \
        self.allDraws, \
        self.drawNames = \
            eb.defineNumberingOfElementaryExpressions(collectionOfFormulas,
                                                      variableNames)

        ### List of tuples (ell,u) containing the lower and upper bounds
        # for each free parameter
        self.bounds = list()
        for x in self.freeBetaNames:
            self.bounds.append((self.allFreeBetas[x].lb,
                                self.allFreeBetas[x].ub))
        ## List of ids of the free beta parameters (those to be estimated)
        self.betaIds = [i for i in range(len(self.freeBetaNames))]

        ## List of initial values of the free beta parameters (those to be estimated)
        self.betaInitValues = [float(self.allFreeBetas[x].initValue) for x in self.freeBetaNames]
        ## Values of the fixed parameters (not estimated).
        self.fixedBetaValues = [float(self.allFixedBetas[x].initValue) for x in self.fixedBetaNames]

        
    def calculateInitLikelihood(self):
        """Calculate the value of the log likelihood function

        The default values of the parameters are used.

        :return: value of the log likelihood.
        :rtype: float.
        """

        ## Value of the loglikelihood for the default values of the parameters.
        self.initLogLike = self.calculateLikelihood(self.betaInitValues,scaled=False)
        return self.initLogLike

    def calculateLikelihood(self, x, scaled, batch=None):
        """Calculates the value of the log likelihood function

        :param x: vector of values for the parameters.
        :type x: list(float)

        :param scaled: if True, the value is diviced by the number of
                       observations used to calculate it. In this
                       case, the values with different sample sizes
                       are comparable. Default: True
        :type scaled: bool

        :param batch: if not None, calculates the likelihood on a
                       random sample of the data. The value of the
                       parameter must be strictly between 0 and 1, and
                       represents the share of the data that will be
                       used. Default: None
        :type batch: float

        :return: the calculated value of the log likelihood
        :rtype: float.

        :raises ValueError: if the length of the list x is incorrect.

        """

        if len(x) != len(self.betaInitValues):
            raise ValueError(f'Input vector must be of length {len(self.betaInitValues)} and not {len(x)}')

        self._prepareDatabaseForFormula(batch)
        f = self.theC.calculateLikelihood(x, self.fixedBetaValues)

        self.logger.general(f"Log likelihood (N={self.database.getSampleSize()}): {f:10.7g}")
        
        if scaled:
            return f / float(self.database.getSampleSize())
        else:
            return f

    def calculateLikelihoodAndDerivatives(self, x, scaled, hessian=False, bhhh=False, batch=None):
        """Calculate the value of the log likelihood function and its derivatives.

        :param x: vector of values for the parameters.
        :type x: list(float)

        :param hessian: if True, the hessian is calculated. Default: False.
        :type hessian: bool

        :param bhhh: if True, the BHHH matrix is calculated. Default: False.
        :type bhhh: bool

        :param batch: if not None, calculates the likelihood on a
                       random sample of the data. The value of the
                       parameter must be strictly between 0 and 1, and
                       represents the share of the data that will be
                       used. Default: None
        :type batch: float


        :return: f, g, h, bh where

                - f is the value of the function (float)
                - g is the gradient (numpy.array)
                - h is the hessian (numpy.array)
                - bh is the BHHH matrix (numpy.array)

        :rtype: tuple  float, numpy.array, numpy.array, numpy.array

        :raises ValueError: if the length of the list x is incorrect

        """

        if len(x) != len(self.betaInitValues):
            raise ValueError(f'Input vector must be of length {len(betaInitValues)} and not {len(x)}')
        self._prepareDatabaseForFormula(batch)
            
        f, g, h, bh = self.theC.calculateLikelihoodAndDerivatives(x,
                                                                  self.fixedBetaValues,
                                                                  self.betaIds,
                                                                  hessian,
                                                                  bhhh)

        if len(self.freeBetaNames) <= 30:
            for i in range(len(self.freeBetaNames)):
                self.logger.debug(f"{self.freeBetaNames[i]}: {x[i]:10.7g}")
        hmsg = ''
        if hessian:
            hmsg = f'Hessian norm:  {np.linalg.norm(h):10.1g}'
        bhhhmsg = ''
        if bhhh:
            bhhhmsg = f'BHHH norm:  {np.linalg.norm(bh):10.1g}'
        self.logger.general(f"Log likelihood (N={self.database.getSampleSize()}): {f:10.7g} Gradient norm: {np.linalg.norm(g):10.1g} {hmsg} {bhhhmsg}")

        if self.saveIterations:
            if self.bestIteration is None:
                self.bestIteration = f
            if f >= self.bestIteration:
                with open(self.fileIterations, 'w') as pf:
                    for i in range(len(x)):
                        print(f"{self.freeBetaNames[i]} = {x[i]}", file=pf)

        if scaled:
            return f / float(self.database.getSampleSize()), np.asarray(g) / float(self.database.getSampleSize()), np.asarray(h) / float(self.database.getSampleSize()), np.asarray(bh) / float(self.database.getSampleSize())
            return f
        else:
            return f, np.asarray(g), np.asarray(h), np.asarray(bh)

    def likelihoodFiniteDifferenceHessian(self, x):
        """Calculate the hessian of the log likelihood function using finite differences.

        May be useful when the analytical hessian has numerical issues.

        :param x: vector of values for the parameters.
        :type x: list(float)

        :return: finite differences approximation of the hessian.
        :rtype: numpy.array

        :raises ValueError: if the length of the list x is incorrect

        """

        def theFunction(x):
            f, g, h, b = self.calculateLikelihoodAndDerivatives(x,
                                                                scaled=False,
                                                                hessian=False,
                                                                bhhh=False)
            return f, np.asarray(g)
        return tools.findiff_H(theFunction, np.asarray(x))

    def checkDerivatives(self, verbose=False):
        """Verifies the implementation of the derivatives.

        It compares the analytical version with the finite differences approximation.

        :param verbose: if True, the comparisons are reported. Default: False.
        :type verbose: bool

        :rtype: tuple.

        :return: f, g, h, gdiff, hdiff where

            - f is the value of the function,
            - g is the analytical gradient,
            - h is the analytical hessian,
            - gdiff is the difference between the analytical and the
              finite differences gradient,
            - hdiff is the difference between the analytical and the
              finite differences hessian,

        """
        def theFunction(x):
            """ Wrapper function to use tools.checkDerivatives """
            f, g, h, b = self.calculateLikelihoodAndDerivatives(x,
                                                                scaled=False,
                                                                hessian=True,
                                                                bhhh=False)
            return f, np.asarray(g), np.asarray(h)

        return tools.checkDerivatives(theFunction,
                                      np.asarray(self.betaInitValues),
                                      self.freeBetaNames,
                                      verbose)


    def loadSavedIteration(self, filename='__savedIterations.txt'):
        """Reads the values of the parameters from a text file where each line
has the form name_of_beta = value_of_beta, and use these values in all
formulas.

        :param filename: name of the text file to read. Default: '__savedIterations.txt'
        :type filename: str. 

        """
        betas = {}
        try:
            with open(filename) as fp:
                for line in fp:
                    l = line.split('=')
                    betas[l[0].strip()] = float(l[1])
            self.changeInitValues(betas)
            logger.detailed(f'Parameter values restored from {filename}')
        except IOError as e:
            logger.warning(f'Cannot read file {filename}. Statement is ignored.')

    def changeInitValues(self, betas):
        if self.loglike is not None:
            self.loglike.changeInitValues(betas)
        if self.weight is not None:
            self.weight.changeInitValues(betas)
        for k, f in self.formulas.items():
            f.changeInitValues(betas)
        for i in range(len(self.freeBetaNames)):
            value = betas.get(self.freeBetaNames[i])
            if value is not None:
                self.betaInitValues[i] = value
            
    def estimate(self,
                 bootstrap=0,
                 algorithm=opt.scipy,
                 algoParameters=None,
                 cfsqpDefaultBounds=1000.0,
                 saveIterations=False,
                 fileIterations='__savedIterations.txt'):

        """Estimate the parameters of the model.

        :param bootstrap: number of bootstrap resampling used to
               calculate the variance-covariance matrix using
               bootstrapping. If the number is 0, bootstrapping is not
               applied. Default: 0.
        :type bootstrap: int

        :param algorithm: optimization algorithm to use for the
               maximum likelihood estimation. If None, cfsqp is
               used. Default: the optimization algorithm available
               with scipy.
        :type algorithm: function

        :param algoParameters: parameters to transfer to the optimization algorithm
        :type algoParameters: dict

        :param cfsqpDefaultBounds: if the user does not provide bounds
              on the parameters, CFSQP assumes that the bounds are
              [-cfsqpDefaultBounds,cfsqpDefaultBounds]
        :type cfsqpDefaultBounds: float

        :param saveIterations: if True, the values of the parameters
                               corresponding to the largest value of
                               the likelihood function are saved in a
                               pickle file at each iteration of the
                               algorithm. Default: False.  
        :type saveIterations: bool

        :param fileIterations: name of the file where to save the
                               values of the parameters. Default:
                               '__savedIterations.txt' 
        :type fileIterations: str

        :return: object containing the estimation results.
        :rtype: biogeme.bioResults

        Example::

            # Create an instance of biogeme
            biogeme  = bio.BIOGEME(database, logprob)
            # Gives a name to the model
            biogeme.modelName = "mymodel"
            # Estimate the parameters
            results = biogeme.estimate()

        :raises biogemeError: if no expression has been provided for the likelihood

        """

        if self.loglike is None:
            raise excep.biogemeError("No log likelihood function has been specificed")
        if len(self.freeBetaNames) == 0:
            raise excep.biogemeError(f"There is no parameter to estimate in the formula: {self.loglike}.")


        self.algorithm = algorithm
        self.algoParameters = algoParameters
        self.cfsqpDefaultBounds = cfsqpDefaultBounds

        self.calculateInitLikelihood()
        self.saveIterations = saveIterations
        self.fileIterations = f'{fileIterations}'
        self.bestIteration = None

        start_time = datetime.now()
        #        yep.start('profile.out')

        #        yep.stop()

        xstar, nIter, nFunctions, diag = self.optimize(self.betaInitValues)
        ## Running time of the optimization algorithm
        self.optimizationTime = datetime.now() - start_time
        ## Information provided by the optimization algorithm after completion.
        self.optimizationMessage = diag
        ## Number of times that the loglikelihood function has been evaluated.
        self.numberOfFunctionEval = nFunctions
        ## Number of iterations performed by the optimization algorithm.
        self.numberOfIterations = nIter
        fgHb = self.calculateLikelihoodAndDerivatives(xstar,
                                                      scaled=False,
                                                      hessian=True,
                                                      bhhh=True)
        if not np.isfinite(fgHb[2]).all():
            self.logger.warning("Numerical problems in calculating the analytical hessian. Finite differences is tried instead.")
            finDiffHessian = self.likelihoodFiniteDifferenceHessian(xstar)
            if not np.isfinite(fgHb[2]).all():
                self.logger.warning("Numerical problems with finite difference hessian as well.")
            else:
                fgHb = fgHb[0], fgHb[1], finDiffHessian, fgHb[3]
        ## numpy array, of size B x K,
        # where
        #        - B is the number of bootstrap iterations
        #        - K is the number pf parameters to estimate
        self.bootstrapResults = None
        if bootstrap > 0:
            start_time = datetime.now()

            self.logger.general(f'Re-estimate the model {bootstrap} times for bootstrapping')
            self.bootstrapResults = np.empty(shape=[bootstrap, len(xstar)])
            for b in range(bootstrap):
                if self.database.isPanel():
                    sample = self.database.sampleIndividualMapWithReplacement()
                    self.theC.setDataMap(sample)
                else:
                    sample = self.database.sampleWithReplacement()
                    self.theC.setData(sample)
                x_br, nIter_br, nFunctions_br, diag_br = self.optimize(xstar)
                self.bootstrapResults[b] = x_br
#            print("*")

            ## Time needed to generate the bootstrap results
            self.bootstrapTime = datetime.now() - start_time

        rawResults = res.rawResults(self,
                                    xstar,
                                    fgHb,
                                    bootstrap=self.bootstrapResults)
        r = res.bioResults(rawResults)
        r.comments = self.comments
        if self.generateHtml:
            r.writeHtml()
        if self.generatePickle:
            r.writePickle()
        return r

    def optimize(self, startingValues):
        """ Calls the optimization algorithm.

        The function self.algorithm is called. If None, CFSQP is invoked.

        :param startingValues: initial values of the parameters.
        :type startingValues: list of floats

        :returns: x, nit, nfev, message, where

               - x is the solution found,
               - nit is the number of iterations performed,
               - nfev is the number of time that the objective function has been evaluated, and
               - message is the diagnostic provided by the algorithm.

        :rtype: tuple
        """

        theFunction = negativeLikelihood(like = self.calculateLikelihood,
                                         likeDeriv = self.calculateLikelihoodAndDerivatives,
                                         scaled = True)
        
        if self.algorithm is None:
            parameters = {'mode':100,
                        'iprint':self.logger.screenLevel,
                        'miter':1000,
                        'eps':6.05545e-06}
            return self.cfsqp(self.betaInitValues,
                              self.fixedBetaValues,
                              self.betaIds,
                              self.bounds,
                              parameters)
        else:
            
            return self.algorithm(theFunction,
                                  self.betaInitValues,
                                  self.fixedBetaValues,
                                  self.betaIds,
                                  self.bounds,
                                  self.algoParameters)


    def simulate(self, theBetaValues=None):
        """Applies the formulas to each row of the database.

        :param theBetaValues: values of the parameters to be used in
                the calculations. If None, the default values are
                used. Default: None.
        :type theBetaValues: dict(str, float) or list(float)

        :return: a pandas data frame with the simulated value. Each
              row corresponds to a row in the database, and each
              column to a formula.

        :rtype: Pandas data frame

        Example::

              # Read the estimation results from a file
              results = res.bioResults(pickleFile='myModel.pickle')
              # Simulate the formulas using the nominal values
              simulatedValues = biogeme.simulate(betaValues)

        :raises biogemeError: if the number of parameters is incorrect

        """

        if theBetaValues is None:
            betaValues = self.betaInitValues
        else:
            if type(theBetaValues) is not dict:
                if len(theBetaValues) != len(self.betaInitValues):
                    err = f"The value of {len(self.betaInitValues)} parameters should be provided, not {len(theBetaValues)}."
                    raise excep.biogemeError(err)

                betaValues = theBetaValues
            else:
                betaValues = list()
                for i in range(len(self.freeBetaNames)):
                    x = self.freeBetaNames[i]
                    if x in theBetaValues:
                        betaValues.append(theBetaValues[x])
                    else:
                        betaValues.append(self.betaInitValues[i])

        output = pd.DataFrame(index=self.database.data.index)
        for k, v in self.formulas.items():
            signature = v.getSignature()
            result = self.theC.simulateFormula(signature,
                                               betaValues,
                                               self.fixedBetaValues,
                                               self.database.data)
            output[k] = result
        return output


    def confidenceIntervals(self, betaValues, intervalSize=0.9):
        """Calculate confidence intervals on the simulated quantities


        :param betaValues: array of parameters values to be used in
               the calculations. Typically, it is a sample drawn from
               a distribution.
        :type betaValues: list(dict(float))

        :param intervalSize: size of the reported confidence interval,
                    in percentage. If it is denoted by s, the interval
                    is calculated for the quantiles (1-s)/2 and
                    (1+s)/2. The default (0.9) corresponds to
                    quantiles for the confidence interval [0.05,0.95].
        :type intervalSize: float

        :return: two pandas data frames 'left' and 'right' with the
            same dimensions. Each row corresponds to a row in the
            database, and each column to a formula. 'left' contains the
            left value of the confidence interval, and 'right' the right
            value

            Example::

                # Read the estimation results from a file
                results = res.bioResults(pickleFile='myModel.pickle')
                # Retrieve the names of the betas parameters that have been estimated
                betas = biogeme.freeBetaNames
                # Draw 100 realization of the distribution of the estimators
                b = results.getBetasForSensitivityAnalysis(betas, size=100)
                # Simulate the formulas using the nominal values
                simulatedValues = biogeme.simulate(betaValues)
                # Calculate the confidence intervals for each formula
                left, right = biogeme.confidenceIntervals(b, 0.9)

        :rtype: tuple of two Pandas dataframes.

        """
        listOfResults = []
        for b in betaValues:
            r = self.simulate(b)
            listOfResults += [r]
        allResults = pd.concat(listOfResults)
        r = (1.0-intervalSize)/2.0
        left = allResults.groupby(level=0).quantile(r)
        right = allResults.groupby(level=0).quantile(1.0-r)
        return left, right

    def createLogFile(self, level=3):
        """ Creates a log file with the messages produced by Biogeme.

        The name of the file is the name of the model with an extension .log

        :param level: types of messages to be captured

            - 0: no output
            - 1: warnings
            - 2: only general information
            - 3: more verbose
            - 4: debug messages

            Default: 3.

        :type level: int

        """
        self.logger.createLog(fileLevel=level, fileName=self.modelName)

    def __str__(self):
        r = f'{self.modelName}: database [{self.database.name}]'
        r += str(self.formulas)
        print(r)
        return r

    def cfsqp(self, betas, fixedBetas, betaIds, bounds, parameters):
        """
        Invokes the CFSQP algorithm for estimation

        :param betas: initial values of the parameters to be estimated.
        :type betas: list of float
        :param fixedBetas: values of the parameters that are not estimated.
        :type fixedBetas: list of float
        :param betaIds: list of IDs of the beta to be estimated.
        :type betaIds: list of int
        :param bounds: lower and upper bounds on each parameter.
        :type bounds: list of tuple
        :param parameters: user defined parameters for CFSQP

           - mode=CBA: specifies job options as described below

                 * A = 0: ordinary minimax problems
                 * A = 1: ordinary minimax problems with each individual function replaced by its absolute value, ie, an L_infty problem
                 * B = 0: monotone decrease of objective function after each iteration
                 * B = 1: monotone decrease of objective function after at most four iterations
                 * C = 1: default operation.
                 * C= 2: requires that constraints always be evaluated before objectives during the line search.

           -  iprint: print level indicator with the following options

               * iprint=0: no normal output, only error information (this option is imposed during phase 1)
               * iprint=1: a final printout at a local solution
               * iprint=2: a brief printout at the end of each iteration
               * iprint=3: detailed infomation is printed out at the end of each iteration (for debugging purposes)

           -  miter: maximum number of iterations allowed by the user to solve              the problem


        :type parameters: dict

        """
        lb, ub = zip(*bounds)
        lb = [-self.cfsqpDefaultBounds if v is None else v for v in lb]
        ub = [self.cfsqpDefaultBounds if v is None else v for v in ub]
        self.theC.setBounds(lb, ub)
        mode = parameters.get('mode', 100)
        iprint = parameters.get('iprint', 2)
        if iprint > 3:
            iprint = 3
        miter = parameters.get('miter', 1000)
        eps = parameters.get('eps', 6.05545e-06)

        output = self.theC.cfsqp(betas,
                                 fixedBetas,
                                 betaIds,
                                 mode,
                                 iprint,
                                 miter,
                                 eps)
        return output

class negativeLikelihood(opt.functionToMinimize):
    """Provides the value of the function to be minimized, as well as its
        derivatives. To be used by the opimization package.
    """

    def __init__(self, like, likeDeriv, scaled):
        """ Constructor
        """
        self.recalculate = True
        self.x = None
        self.batch = None
        self.fv = None
        self.gv = None
        self.hv = None
        self.bhhhv = None
        self.like = like
        self.likeDeriv = likeDeriv
        self.scaled = scaled


    def setVariables(self,x):
        self.recalculate = True
        self.x = x
        self.fv = None
        self.gv = None
        self.hv = None
        self.bhhhv = None

    def f(self, batch = None):
        if self.x is None:
            raise excep.biogemeError("The variables must be set first.")

        if batch is not None or self.batch is not None:
            self.batch = batch
            self.recalculate = True

        if self.fv is None:
            self.recalculate = True

        if self.recalculate:
            self.fv = self.like(self.x, self.scaled, self.batch)
            self.gv = None
            self.hv = None
            self.bhhhv = None

        return -self.fv

    def f_g(self, batch = None):
        if self.x is None:
            raise excep.biogemeError("The variables must be set first.")

        if batch is not None or self.batch is not None:
            self.batch = batch
            self.recalculate = True

        if self.fv is None or self.gv is None:
            self.recalculate = True

        if self.recalculate:
            self.fv, self.gv, *_ = self.likeDeriv(self.x, self.scaled, hessian=False, bhhh=False, batch=batch)
            self.hv = None
            self.bhhhv = None

        return -self.fv, -self.gv

    def f_g_h(self, batch=None):
        if self.x is None:
            raise excep.biogemeError("The variables must be set first.")

        if batch is not None or self.batch is not None:
            self.batch = batch
            self.recalculate = True

        if self.fv is None or self.gv is None or self.hv is None:
            self.recalculate = True

        if self.recalculate:
            self.fv, self.gv, self.hv, _ = self.likeDeriv(self.x, self.scaled, hessian=True, bhhh=False, batch=batch)
            self.bhhhv = None

        return -self.fv, -self.gv, -self.hv

    def f_g_bhhh(self, batch = None):
        if batch is not None or self.batch is not None:
            self.batch = batch
            self.recalculate = True

        if self.x is None:
            raise excep.biogemeError("The variables must be set first.")

        if self.fv is None or self.gv is None or self.bhhhv is None:
            self.recalculate = True

        if self.recalculate:
            self.fv, self.gv, _ , self.bhhhv = self.likeDeriv(self.x, self.scaled,hessian=False, bhhh=True, batch=batch)
            self.hv = None

        return -self.fv, -self.gv, -self.bhhhv

    



# distutils: language=c++
# cython: embedsignature=True

cimport numpy as np
import numpy as np
from libcpp.vector cimport vector
from libcpp.string cimport string
from libcpp cimport bool as bool_t

ctypedef vector[unsigned long] uint_vector
ctypedef vector[uint_vector] uint_matrix
ctypedef vector[double] double_vector
ctypedef vector[double_vector] double_matrix
ctypedef vector[double_matrix] double_tensor


cdef extern from "biogeme.h":

	cdef cppclass biogeme:
		biogeme() except +

		double calculateLikelihood(double_vector betas, 
			double_vector fixedBetas) except +

		double calculateLikeAndDerivatives(double_vector betas, 
			double_vector fixedBetas, 
			uint_vector betaIds, 
			double_vector& g,
			double_matrix& h,
			double_matrix& bhhh,
			bool_t hessian,
			bool_t bhhh) except +

		string cfsqp(double_vector betas,
			double_vector fixedBetas,
			uint_vector betaIds,
			unsigned long nit,
			unsigned long nf,
			unsigned long mode,
			unsigned long iprint,
			unsigned long miter,
			double eps) except +

		void setPanel(bool_t p)

		void setBounds(double_vector lb, double_vector ub)

		void simulateFormula(vector[string] loglikeSignatures,
					double_vector betas, 
					double_vector fixedBetas,
				     double_matrix& data,
				     double_vector& results) except +

		void setExpressions(vector[string] loglikeSignatures, 
						vector[string] weightSignatures,
						unsigned long numberOfThreads)

		void setData(double_matrix& d)

		void setDataMap(uint_matrix& dm)

		void setMissingData(double md)
		
		void setDraws(double_tensor& draws)


cdef class pyBiogeme:
	cdef biogeme theBiogeme

	def __cinit__(self):
		self.theBiogeme = biogeme()

	def setPanel(self,panel=True):
		self.theBiogeme.setPanel(panel)

	def calculateLikelihoodAndDerivatives(self,betas,fixedBetas,betaIds,hessian,bhhh,draws=None):
		cdef double_vector g
		cdef double_matrix h
		cdef double_matrix b
		g = np.empty(len(betas))
		h = np.empty([len(betas),len(betas)])
		b = np.empty([len(betas),len(betas)])
		f = self.theBiogeme.calculateLikeAndDerivatives(betas,fixedBetas,betaIds,g,h,b,hessian,bhhh)
		return f,g,h,b

	def cfsqp(self,betas,fixedBetas,betaIds,mode,iprint,moter,eps):
		cdef double_vector b = betas
		cdef unsigned long nit
		cdef unsigned long nf
		diag = self.theBiogeme.cfsqp(b,fixedBetas,betaIds,nit,nf,mode,iprint,moter,eps)
		return b,nit,nf,diag

	def setBounds(self,lb,ub):
		self.theBiogeme.setBounds(lb,ub)

	def calculateLikelihood(self, betas,fixedBetas):
		r = self.theBiogeme.calculateLikelihood(betas,fixedBetas)
		return r

	def simulateFormula(self, formula, betas,fixedBetas, d):	
		cdef double_vector r
		d = np.ascontiguousarray(d)
		self.theBiogeme.simulateFormula(formula,betas,fixedBetas,d,r)
		return r
	
	def setExpressions(self,loglikeFormulas,nbrOfThreads,weightFormulas=None):
		cdef vector[string] w
		if (weightFormulas is not None):
			w = weightFormulas
		self.theBiogeme.setExpressions(loglikeFormulas,w,nbrOfThreads)

	def setData(self, d):
		d = np.ascontiguousarray(d)
		self.theBiogeme.setData(d)

	def setDataMap(self, m):
		m = np.ascontiguousarray(m)
		self.theBiogeme.setDataMap(m)

	def setMissingData(self, md):
		self.theBiogeme.setMissingData(md)


	def setDraws(self, draws):
		draws = np.ascontiguousarray(draws)
		self.theBiogeme.setDraws(draws)

				


				





""" File: 03antithetic.py

 Author: Michel Bierlaire, EPFL
 Date: Wed Dec 11 17:00:05 2019

Calculation of a simple integral using Monte-Carlo integration with
Halton draws base 13. It illustrates how to use draws that
are not directly available in Biogeme.


"""

import numpy as np
import pandas as pd
import biogeme.database as db
import biogeme.biogeme as bio
import biogeme.draws as draws
from biogeme.expressions import exp, bioDraws, MonteCarlo


# We create a fake database with one entry, as it is required to store the draws
pandas = pd.DataFrame()
pandas['FakeColumn'] = [1.0]
database = db.Database("fakeDatabase",pandas)

# The user can define new draws. For example, Halton draws with base 13, skipping the first 10 draws.

def halton13_anti(sampleSize,numberOfDraws):
    R = int(numberOfDraws / 2)
    d = draws.getHaltonDraws(sampleSize,
                                 R,
                                 base=13,
                                 skip=10)
    return np.concatenate((d,1-d),axis=1)

mydraws = {'HALTON13_ANTI':(halton13_anti,'Antithetic Halton draws, base 13, skipping 10')}
database.setRandomNumberGenerators(mydraws)


integrand = exp(bioDraws('U','UNIFORM_ANTI'))
simulatedI = MonteCarlo(integrand)

integrand_halton13 = exp(bioDraws('U_halton13','HALTON13_ANTI'))
simulatedI_halton13 = MonteCarlo(integrand_halton13)

integrand_mlhs = exp(bioDraws('U_mlhs','UNIFORM_MLHS_ANTI'))
simulatedI_mlhs = MonteCarlo(integrand_mlhs)

trueI = exp(1.0) - 1.0 

R = 20000

error = simulatedI - trueI

error_halton13 = simulatedI_halton13 - trueI

error_mlhs = simulatedI_mlhs - trueI


simulate = {'Analytical Integral': trueI,
            'Simulated Integral': simulatedI,
            'Error             ': error,
            'Simulated Integral (Halton13)': simulatedI_halton13,
            'Error (Halton13)             ': error_halton13,
            'Simulated Integral (MLHS)': simulatedI_mlhs,
            'Error (MLHS)             ': error_mlhs}



biogeme  = bio.BIOGEME(database,simulate,numberOfDraws=R)
biogeme.modelName = "03antithetic"
results = biogeme.simulate()
print(f"Analytical integral:{results.iloc[0]['Analytical Integral']:.6f}")
print(f"\t\tUniform (Anti)\t\tHalton13 (Anti)\t\tMLHS (Anti)")
print(f"Simulated\t{results.iloc[0]['Simulated Integral']:.6g}\t{results.iloc[0]['Simulated Integral (Halton13)']:.6g}\t{results.iloc[0]['Simulated Integral (MLHS)']:.6g}")
print(f"Error\t\t{results.iloc[0]['Error             ']:.6g}\t{results.iloc[0]['Error (Halton13)             ']:.6g}\t{results.iloc[0]['Error (MLHS)             ']:.6g}")





"""
Implementation of class contaning and processing the estimation results.

:author: Michel Bierlaire
:date: Tue Mar 26 16:50:01 2019

"""
import pickle
import datetime
import pandas as pd
import numpy as np
from scipy import linalg
from scipy import stats
import biogeme.version as bv
import biogeme.filenames as bf
import biogeme.exceptions as excep
import biogeme.messaging as msg



def calcPValue(t):
    """ Calculates the p value of a parameter from its t-statistic.

    The formula is

    .. math:: 2(1-\\Phi(|t|)

    where :math:`\Phi(\\cdot)` is the CDF of a normal distribution.

    :param t: t-statistics
    :type: float

    :return: p-value
    :rtype: float
    """
    p = 2.0 * (1.0 - stats.norm.cdf(abs(t)))
    return p

class beta:
    """ Class gathering the information related to the parameters of the model
    """
    def __init__(self,name,value,bounds):
        """
        Constructor

        :param name: name of the parameter.
        :type name: string
        :param value: value of the parameter.
        :type value: float
        :param bounds: tuple (l,b) with lower and upper bounds
        :type bounds: float,float
        """
        ## Name of the parameter
        self.name = name
        ## Current value
        self.value = value
        ## Lower bound
        self.lb = bounds[0]
        ## Upper bound
        self.ub = bounds[1]
        ## Standard error
        self.stdErr = None
        ## t-test
        self.tTest = None
        ## p-value
        self.pValue = None
        ## Robust standard error
        self.robust_stdErr = None
        ## Robust t-test
        self.robust_tTest = None
        ## Robust p-value
        self.robust_pValue = None
        ## Standard error calculated from bootstrap
        self.bootstrap_stdErr = None
        ## t-test  calculated from bootstrap
        self.bootstrap_tTest = None
        ## p-value calculated from bootstrap
        self.bootstrap_pValue = None
        

    def isBoundActive(self,threshold=1.0e-6):
        """Check if one of the two bound is "numerically" active.

        Being numerically active means that the distance between the value of the parameter and one of its bounds is below the threshold.
        
        :param threshold: distance below which the bound is considered to be active. Default: :math:`10^{-6}`

        :return: True is one of the two bounds is numericall y active.
        :rtype: bool

        """
        if threshold < 0:
            raise excep.biogemeError(f"Threshold ({threshold}) must be non negative")
            
        if self.lb is not None and np.abs(self.value - self.lb) <= threshold:
            return True
        if self.ub is not None and np.abs(self.value - self.ub) <= threshold:
            return True
        return False
    
    def setStdErr(self,se):
        """ Records the standard error, and calculates and records 
            the corresponding t-statistic and p-value
        
        :param se: standard error.
        :type se: float

        """
        self.stdErr = se
        if se == 0:
            self.tTest = np.finfo(float).max
        else:
            self.tTest = np.nan_to_num(self.value / se)
        self.pValue = calcPValue(self.tTest)

    def setRobustStdErr(self,se):
        """ Records the robust standard error, and calculates and records 
            the corresponding t-statistic and p-value

        
        :param se: robust standard error
        :type se: float
        
        """
        self.robust_stdErr = se
        if se == 0:
            self.robust_tTest = np.finfo(float).max
        else:
            self.robust_tTest = np.nan_to_num(self.value / se)
        self.robust_pValue = calcPValue(self.robust_tTest)

    def setBootstrapStdErr(self,se):
        """ Records the robust standard error calculated by bootstrap, and calculates and records the corresponding t-statistic and p-value

        :param se: standard error calculated by bootstrap.
        :type se: float
        """
        self.bootstrap_stdErr = se
        if se == 0:
            self.bootstrap_tTest = np.finfo(float).max
        else:
            self.bootstrap_tTest = np.nan_to_num(self.value / se)
        self.bootstrap_pValue = calcPValue(self.robust_tTest)
        
        
    def __str__(self):
        str = "{:15}: {:.3g}".format(self.name,self.value)
        if self.stdErr is not None:
            str += "[{:.3g} {:.3g} {:.3g}]".format(self.stdErr,self.tTest,self.pValue)
        if self.robust_stdErr is not None:
            str += "[{:.3g} {:.3g} {:.3g}]".format(self.robust_stdErr,self.robust_tTest,self.robust_pValue)
        if self.bootstrap_stdErr is not None:
            str += "[{:.3g} {:.3g} {:.3g}]".format(self.bootstrap_stdErr,self.bootstrap_tTest,self.bootstrap_pValue)
        return str

class rawResults:
    """ Class containing the raw results from the estimation
    """
    def __init__(self,theModel,betaValues,fgHb,bootstrap=None):
        """
        Constructor

        :param theModel: object with the model
        :type theModel: biogeme.BIOGEME
        :param betaValues: list containing the estimated values of the parameters
        :type betaValues: list(float)
        :param fgHb: tuple f,g,H,bhhh containing

                 - f: the value of the function, 
                 - g: the gradient,
                 - H: the second derivative matrix,
                 - bhhh: the BHHH matrix.
        :type fgHb: float,numpy.array, numpy.array, numpy.array

        :param bootstrap: output of the bootstrapping. numpy array, of size B x K,  where

                - B is the number of bootstrap iterations
                - K is the number of parameters to estimate

                          Default: None.
        :type bootstrap: numpy.array
    """

        ## Name of the model
        self.modelName = theModel.modelName
        ## Number of parameters
        self.nparam = len(betaValues)
        ## Values of the parameters
        self.betaValues = betaValues
        ## Names of the parameters
        self.betaNames = theModel.freeBetaNames
        ## Value of the likelihood function with the initial value of the parameters
        self.initLogLike = theModel.initLogLike
        ## List of objects of type results.beta
        self.betas = list()
        for b,n in zip(betaValues,self.betaNames):
            bounds = theModel.getBoundsOnBeta(n)
            self.betas.append(beta(n,b,bounds))
        ## Value of the loglikelihood function
        self.logLike = fgHb[0]
        ## Value of the gradient of the loglikelihood function
        self.g = fgHb[1]
        ## Value of the hessian of the loglikelihood function
        self.H = fgHb[2]
        ## Value of the BHHH matrix of the loglikelihood function
        self.bhhh = fgHb[3]
        ## Name of the database
        self.dataname = theModel.database.name
        ## Sample size (number of individuals if panel data)
        self.sampleSize = theModel.database.getSampleSize()
        ## NUmber of observations
        self.numberOfObservations = theModel.database.getNumberOfObservations()
        ## True if the model involved Monte Carlo integration
        self.monteCarlo = theModel.monteCarlo
        ## Number of draws for Monte Carlo integration
        self.numberOfDraws = theModel.numberOfDraws
        ## Types of draws for Monte Carlo integration
        self.typesOfDraws = theModel.database.typesOfDraws
        ## Number of excluded data
        self.excludedData = theModel.database.excludedData
        ## Time needed to process the draws
        self.drawsProcessingTime = theModel.drawsProcessingTime
        ## Time needed for estimation of the parameters
        self.optimizationTime = theModel.optimizationTime
        ## Norm of the gradient
        self.gradientNorm = linalg.norm(self.g)
        ## Diagnostic given by the optimization algorithm
        self.optimizationMessage = theModel.optimizationMessage
        ## Number of evaluations of the function
        self.numberOfFunctionEval = theModel.numberOfFunctionEval
        ## Number of iterations of the algorithm
        self.numberOfIterations = theModel.numberOfIterations
        ## Number of threads used for parallel computing
        self.numberOfThreads = theModel.numberOfThreads
        ## Name of the HTML output file
        self.htmlFileName = None
        ## Name of the LaTeX output file
        self.latexFileName = None
        ## Name of the pickle outpt file
        self.pickleFileName = None
        ##  output of the bootstrapping. numpy array, of size B x K,
        #    where
        #        - B is the number of bootstrap iterations
        #        - K is the number of parameters to estimate
        self.bootstrap = bootstrap
        if bootstrap is not None:
            ## Time needed to perform the bootstrap
            self.bootstrapTime = theModel.bootstrapTime
    
class bioResults:
    """ Class managing the estimation results
    """
    def __init__(self,rawResults=None,pickleFile=None):
        """ Constructor

        :param rawResults: object with the results of the estimation. Default: None.
        :type rawResults: biogeme.results.rawResults 
        :param pickleFile: name of the file containing the raw results in pickle format. Default: None.
        :type pickleFile: string

        :raise biogeme.exceptions.biogemeError: if no data is provided.
        """

        self.logger = msg.bioMessage()
        
        if rawResults is not None:
            ### Object of type results.rawResults contaning the raw estimation results.
            self.data = rawResults
        elif pickleFile is not None:
            with open(pickleFile, 'rb') as f:
                self.data = pickle.load(f)
        else:
            raise excep.biogemeError("No data provided.")

        self.comments = None
        self._calculateStats()

    def writePickle(self):
        """ Dump the data in a file in pickle format.

        :return: name of the file.
        :rtype: string
        """ 
        self.data.pickleFileName = bf.getNewFileName(self.data.modelName,"pickle")
        with open(self.data.pickleFileName, 'wb') as f:
            pickle.dump(self.data, f)

        self.logger.general(f"Results saved in file {self.data.pickleFileName}")
        return self.data.pickleFileName


        
    def _calculateTest(self,i,j,matrix):
        """ Calculates a t-test comparing two coefficients

        Args:
           i: index of first coefficient \f$\beta_i\f$.
           j: index of second coefficient \f$\beta_i\f$.
           matrix: estimate of the variance-covariance matrix \f$m\f$.

        Returns:
           t test \f[ \frac{\beta_i-\beta_j}{\sqrt{m_{ii} + m_{jj} - 2 m_{ij} }}\f]
        """
        vi = self.data.betaValues[i]
        vj = self.data.betaValues[j]
        varI = matrix[i,i]
        varJ = matrix[j,j]
        covar = matrix[i,j]
        r =  varI + varJ - 2.0 * covar
        if r <= 0:
            test = np.finfo(float).max
        else:
            test = (vi - vj) / np.sqrt(r)
        return test
    
    def _calculateStats(self):
        """Calculates the following statistics:
            - likelihood ratio test between the initial and ethe estimated models: \f[-2(L_0-L^*) \f]
            - Rho square: \f[ 1 - \frac{L^*}{L^0}\f]
            - Rho bar square: \f[ 1 - \frac{L^* - K}{L^0}\f]
            - AIC: \f[2(K - L^*) \f]
            - BIC: \f[ -2 L^* + K  \log(N) \f]

        Estimates for the variance-covariance matrix (Rao-Cramer,
        robust, and bootstrap) are also calculated, as well as t-tests and
        p value for the comparison of pairs of coefficients.

        """
        self.data.likelihoodRatioTest = -2.0 * (self.data.initLogLike - self.data.logLike)
        self.data.rhoSquare = np.nan_to_num(1.0 - self.data.logLike / self.data.initLogLike)
        self.data.rhoBarSquare = np.nan_to_num(1.0 - (self.data.logLike-self.data.nparam) / self.data.initLogLike)
        self.data.akaike = 2.0 * self.data.nparam - 2.0 * self.data.logLike
        self.data.bayesian = - 2.0 * self.data.logLike + self.data.nparam * np.log(self.data.sampleSize) 
        # We calculate the eigenstructure to report in case of singularity
        self.data.eigenValues,self.data.eigenvectors = linalg.eigh(-np.nan_to_num(self.data.H))
        U, self.data.singularValues,V = linalg.svd(-np.nan_to_num(self.data.H))
        # We use the pseudo inverse in case the matrix is singular
        self.data.varCovar = -linalg.pinv(np.nan_to_num(self.data.H))
        for i in range(self.data.nparam):
            if self.data.varCovar[i,i] < 0:
                self.data.betas[i].setStdErr(np.finfo(float).max)
            else:
                self.data.betas[i].setStdErr(np.sqrt(self.data.varCovar[i,i]))

        d = np.diag(self.data.varCovar)
        if (d > 0).all():
            diag = np.diag(np.sqrt(d))
            diagInv = linalg.inv(diag)
            self.data.correlation = diagInv.dot(self.data.varCovar.dot(diagInv))
        else:
            self.data.correlation = np.full_like(self.data.varCovar,np.finfo(float).max)

        
        # Robust estimator
        self.data.robust_varCovar = self.data.varCovar.dot(self.data.bhhh.dot(self.data.varCovar))
        for i in range(self.data.nparam):
            if self.data.robust_varCovar[i,i] < 0:
                self.data.betas[i].setRobustStdErr(np.finfo(float).max)
            else:
                self.data.betas[i].setRobustStdErr(np.sqrt(self.data.robust_varCovar[i,i]))
        rd = np.diag(self.data.robust_varCovar)
        if (rd > 0).all():
            diag = np.diag(np.sqrt(rd))
            diagInv = linalg.inv(diag)
            self.data.robust_correlation = diagInv.dot(self.data.robust_varCovar.dot(diagInv))
        else:
            self.data.robust_correlation = np.full_like(self.data.robust_varCovar,np.finfo(float).max)
            
        # Bootstrap
        if self.data.bootstrap is not None:
            self.data.bootstrap_varCovar = np.cov(self.data.bootstrap,rowvar=False)
            for i in range(self.data.nparam):
                if self.data.bootstrap_varCovar[i,i] < 0:
                    self.data.betas[i].setBootstrapStdErr(np.finfo(float).max)
                else:
                    self.data.betas[i].setBootstrapStdErr(np.sqrt(self.data.bootstrap_varCovar[i,i]))
            rd = np.diag(self.data.bootstrap_varCovar)
            if (rd > 0).all():
                diag = np.diag(np.sqrt(rd))
                diagInv = linalg.inv(diag)
                self.data.bootstrap_correlation = diagInv.dot(self.data.bootstrap_varCovar.dot(diagInv))
            else:
                self.data.bootstrap_correlation = np.full_like(self.data.bootstrap_varCovar,np.finfo(float).max)

        self.data.secondOrderTable = dict()
        for i in range(self.data.nparam):
            for j in range(i):
                t = self._calculateTest(i,j,self.data.varCovar)
                p = calcPValue(t)
                trob = self._calculateTest(i,j,self.data.robust_varCovar)
                prob = calcPValue(trob)
                if self.data.bootstrap is not None:
                    tboot = self._calculateTest(i,j,self.data.bootstrap_varCovar)
                    pboot = calcPValue(tboot)
                name = (self.data.betaNames[i],self.data.betaNames[j])
                if self.data.bootstrap is not None:
                    self.data.secondOrderTable[name] = [self.data.varCovar[i,j],self.data.correlation[i,j],t,p,self.data.robust_varCovar[i,j],self.data.robust_correlation[i,j],trob,prob,self.data.bootstrap_varCovar[i,j],self.data.bootstrap_correlation[i,j],tboot,pboot]
                else:
                    self.data.secondOrderTable[name] = [self.data.varCovar[i,j],self.data.correlation[i,j],t,p,self.data.robust_varCovar[i,j],self.data.robust_correlation[i,j],trob,prob]

        self.data.smallestEigenValue = min(self.data.eigenValues)
        self.data.smallestSingularValue = min(self.data.singularValues)

    def __str__(self):
        r = "\n"
        r += "Results for model ["+self.data.modelName+"]\n"
        if self.data.htmlFileName is not None :
            r += "Output file (HTML):\t\t\t{}\n".format(self.data.htmlFileName)
        if self.data.latexFileName is not None :
            r += "Output file (LaTeX):\t\t\t{}\n".format(self.data.latexFileName)
        r += "Nbr of parameters:\t\t{}\n".format(self.data.nparam)
        r += "Sample size:\t\t\t{}\n".format(self.data.sampleSize)
        if self.data.sampleSize != self.data.numberOfObservations:
            r += "Observations:\t\t\t{}\n".format(self.data.numberOfObservations)
        r += "Excluded data:\t\t\t{}\n".format(self.data.excludedData)
        r += "Init log likelihood:\t\t{:.7g}\n".format(self.data.initLogLike)
        r += "Final log likelihood:\t\t{:.7g}\n".format(self.data.logLike)
        r += "Likelihood ratio test:\t\t{:.7g}\n".format(self.data.likelihoodRatioTest)
        r += "Rho square:\t\t\t{:.3g}\n".format(self.data.rhoSquare)
        r += "Rho bar square:\t\t\t{:.3g}\n".format(self.data.rhoBarSquare)
        r += "Akaike Information Criterion:\t{:.7g}\n".format(self.data.akaike)
        r += "Bayesian Information Criterion:\t{:.7g}\n".format(self.data.bayesian)
        r += "Final gradient norm:\t\t{:.7g}\n".format(self.data.gradientNorm)
        r += "\n".join(["{}".format(b) for b in self.data.betas])
        r += "\n"
        for k,v in self.data.secondOrderTable.items():
            r += "{}:\t{:.3g}\t{:.3g}\t{:.3g}\t{:.3g}\t{:.3g}\t{:.3g}\t{:.3g}\t{:.3g}\n".format(k,*v)
        return r

    def _getLaTeXHeader(self):
        """Prepare the header for the LaTeX file, containing comments and the
        version of Biogeme.

        Return:
           string containing the header.
        """
        h = ""
        h += "%% This file is designed to be included into a LaTeX document\n"
        h += "%% See http://www.latex-project.org/ for information about LaTeX\n"
        h += bv.getLaTeX()
        return h
    
    def getLaTeX(self):
        """ Get the results coded in LaTeX

        :return: LaTeX code
        :rtype: string
        """
        now = datetime.datetime.now()
        h = self._getLaTeXHeader()
        if self.data.latexFileName is not None:
            h += "\n%% File "+self.data.latexFileName
        h += "\n%% This file has automatically been generated on {}</p>\n".format(now)
        if self.data.dataname is not None:
            h += "\n%%Database name: "+self.data.dataname+"\n"
        h += "\n%% General statistics\n"
        h += "\\section{General statistics}\n"
        d = self.getGeneralStatistics()
        h += "\\begin{tabular}{ll}\n"
        for k,(v,p) in d.items():
            if isinstance(v,bytes):
                v = str(v)
            if isinstance(v, str):
                v = v.replace('_','\\_')
            h += f"{k} & {v:{p}} \\\\\n"
        h += "\\end{tabular}\n"

        h += "\n%%Parameter estimates\n"
        h += "\section{Parameter estimates}\n"
        table = self.getEstimatedParameters()

        def formatting(x):
            """ Defines the formatting for the to_latex function of pandas
            """
            res = f"{x:.3g}"
            if '.' in res:
                return res
            else:
                return f"{res}.0"
        
        h += table.to_latex(float_format=formatting)

        h += "\n%%Correlation\n"
        h += "\section{Correlation}\n"
        table = self.getCorrelationResults()
        h += table.to_latex(float_format=formatting)
        return h

    def getGeneralStatistics(self):
        """Format the results in a dict
        
        :return: dict with the results. The keys describe each
                 content. Each element is a tuple, with the value and its
                 preferred formatting. 
    
        Example::

                     'Init log likelihood': (-115.30029248549191, '.7g')

        :rtype: dict(string:float,string)
        """
        d = {}
        d['Number of estimated parameters'] = self.data.nparam,''
        d['Sample size'] = self.data.sampleSize,''
        if self.data.sampleSize != self.data.numberOfObservations:
            d['Observations'] = self.data.numberOfObservations,''
        d['Excluded observations'] = self.data.excludedData,''
        d['Init log likelihood'] = self.data.initLogLike,'.7g'
        d['Final log likelihood'] = self.data.logLike,'.7g'
        d['Likelihood ratio test for the init. model'] = self.data.likelihoodRatioTest,'.7g'
        d['Rho-square for the init. model'] = self.data.rhoSquare,'.3g'
        d['Rho-square-bar for the init. model'] = self.data.rhoBarSquare,'.3g'
        d['Akaike Information Criterion'] = self.data.akaike,'.7g'
        d['Bayesian Information Criterion'] = self.data.bayesian,'.7g'
        d['Final gradient norm'] = self.data.gradientNorm,'.4E'
        d['Diagnostic'] = self.data.optimizationMessage,''
        d['Database readings'] = self.data.numberOfFunctionEval,''
        d['Iterations'] = self.data.numberOfIterations,''
        if self.data.monteCarlo:
            d['Number of draws'] = self.data.numberOfDraws,''
            d['Draws generation time'] = self.data.drawsProcessingTime,''
            d['Types of draws'] = [f"{i}: {k}]" for i,k in self.data.typesOfDraws.items()],''
        d['Optimization time'] = self.data.optimizationTime,''
        if self.data.bootstrap is not None:
            d['Bootstrapping time'] = self.data.bootstrapTime,''
        d['Nbr of threads'] = self.data.numberOfThreads,''
        return d

    def getEstimatedParameters(self):
        """Gather the estimated parameters and the corresponding statistics in
a Pandas dataframe.

        :return: Pandas dataframe with the results
        :rtype: pandas.DataFrame
        
        """
        ### There should be a more 'Pythonic' way to do this.
        anyActiveBound = False
        for b in self.data.betas:
            if b.isBoundActive():
                anyActiveBound = True
        if anyActiveBound:
            columns = ['Value','Active bound','Std err','t-test','p-value','Rob. Std err','Rob. t-test','Rob. p-value']
        else:
            columns = ['Value','Std err','t-test','p-value','Rob. Std err','Rob. t-test','Rob. p-value']
        if self.data.bootstrap is not None:
            columns += [f'Bootstrap[{len(self.data.bootstrap)}] Std err','Bootstrap t-test','Bootstrap p-value']
        table = pd.DataFrame(columns=columns)
        for b in self.data.betas:
            if anyActiveBound:
                arow = {'Value':b.value,
                        'Active bound':{True:1.0,False:0.0}[b.isBoundActive()],
                        'Std err':b.stdErr,
                        't-test':b.tTest,
                        'p-value':b.pValue,
                        'Rob. Std err':b.robust_stdErr,
                        'Rob. t-test':b.robust_tTest,
                        'Rob. p-value':b.robust_pValue}
            else:
                arow = {'Value':b.value,
                        'Std err':b.stdErr,
                        't-test':b.tTest,
                        'p-value':b.pValue,
                        'Rob. Std err':b.robust_stdErr,
                        'Rob. t-test':b.robust_tTest,
                        'Rob. p-value':b.robust_pValue}
            if self.data.bootstrap is not None:
                arow[f'Bootstrap[{len(self.data.bootstrap)}] Std err'] = b.bootstrap_stdErr
                arow['Bootstrap t-test'] = b.bootstrap_tTest
                arow['Bootstrap p-value'] = b.bootstrap_pValue
            
            table.loc[b.name] = pd.Series(arow)
        return table

    def getCorrelationResults(self):
        """ Get the statistics about pairs of coefficients as a Pandas dataframe

        :return: Pandas data frame with the correlation results
        :rtpye: pandas.DataFrame
        """
        columns = ['Covariance','Correlation','t-test','p-value','Rob. cov.','Rob. corr.','Rob. t-test','Rob. p-value']
        if self.data.bootstrap is not None:
            columns += ['Boot. cov.','Boot. corr.','Boot. t-test','Boot. p-value']
        table = pd.DataFrame(columns=columns)
        for k,v in self.data.secondOrderTable.items():
            arow = {'Covariance':v[0],
                    'Correlation':v[1],
                    't-test':v[2],
                    'p-value':v[3],
                    'Rob. cov.':v[4],
                    'Rob. corr.':v[5],
                    'Rob. t-test':v[6],
                    'Rob. p-value':v[7]}
            if self.data.bootstrap is not None:
                arow['Boot. cov.'] = v[8]
                arow['Boot. corr.'] = v[9]
                arow['Boot. t-test'] = v[10]
                arow['Boot. p-value'] = v[11]
            table.loc[f'{k[0]}-{k[1]}'] = pd.Series(arow)
        return table
       
    def getHtml(self):
        """ Get the results coded in HTML

        :return: HTML code
        :rtpye: string
        """
        now = datetime.datetime.now()
        h = self._getHtmlHeader()
        h += bv.getHtml()
        h += "<p>This file has automatically been generated on {}</p>\n".format(now)
        h += "<p>If you drag this HTML file into the Calc application of <a href='http://www.openoffice.org/' target='_blank'>OpenOffice</a>, or the spreadsheet of <a href='https://www.libreoffice.org/' target='_blank'>LibreOffice</a>, you will be able to perform additional calculations.</p>\n"
        h += "<table>\n"
        h += "<tr class=biostyle><td align=right><strong>Report file</strong>:	</td><td>"+self.data.htmlFileName+"</td></tr>\n"
        h += "<tr class=biostyle><td align=right><strong>Database name</strong>:	</td><td>"+self.data.dataname+"</td></tr>\n"
        h += "</table>\n"

        if self.comments is not None:
            h += "<h2>Comments</h2>"
            h += f'<p>{self.comments}</p>'
            
        ### Include here the part on statistics

        h += "<h1>Estimation report</h1>\n"

        h += "<table border='0'>\n"
        d = self.getGeneralStatistics()
        # k is the description of the quantity
        # v is the value
        # p is the precision to format it
        for k,(v,p) in d.items():
            h += f"<tr class=biostyle><td align=right ><strong>{k}</strong>: </td> <td>{v:{p}}</td></tr>\n"
        h += "</table>\n"

        table = self.getEstimatedParameters()

        h += "<h1>Estimated parameters</h1>\n"
        h += "<p><font size='-1'>Click on the headers of the columns to sort the table  [<a href='http://www.kryogenix.org/code/browser/sorttable/' target='_blank'>Credits</a>]</font></p>\n"
        h += "<table border='1' class='sortable'>\n"
        h += "<tr class=biostyle><th>Name</th>"
        for c in table.columns:
            h += f"<th>{c}</th>"
        h += "</tr>\n" 
        for name, values in table.iterrows():
            h += f"<tr class=biostyle><td>{name}</td>"
            for k,v in values.items():
#                print(f"values[{k}] = {v}")
                h += f"<td>{v:.3g}</td>"
            h += "</tr>\n"
        h += "</table>\n"

        table = self.getCorrelationResults()
        h += "<h2>Correlation of coefficients</h2>\n"
        h += "<p><font size='-1'>Click on the headers of the columns to sort the table [<a href='http://www.kryogenix.org/code/browser/sorttable/' target='_blank'>Credits</a>]</font></p>\n"
        h += "<table border='1' class='sortable'>\n"
        h += "<tr class=biostyle><th>Coefficient1</th><th>Coefficient2</th>"
        for c in table.columns:
            h += f"<th>{c}</th>"
        h += "</tr>\n" 
        for name, values in table.iterrows():
            n = name.split('-')
            h += f"<tr class=biostyle><td>{n[0]}</td><td>{n[1]}</td>"
            for k,v in values.items():
               h += f"<td>{v:.3g}</td>"
            h += "</tr>\n"
        h += "</table>\n"
                
        h += "<p>Smallest eigenvalue: {:.6g}</p>\n".format(self.data.smallestEigenValue)
        h += "<p>Smallest singular value: {:.6g}</p>\n".format(self.data.smallestSingularValue)
        h += self._getHtmlFooter()
        return h

    def getBetaValues(self,myBetas=None):
        """Retrieve the values of the estimated parameters, by names.
        
        :param myBetas: names of the requested parameters. If None, all
                  available parameters will be reported. Default: None.
        :type myBetas: list(string)

        :return: dict containing the values, where the keys are the names.
        :rtype: dict(string:float)


        :raise biogeme.exceptions.biogemeError: if some requested parameters are not available.
        """
        values = dict()
        if myBetas is None:
            myBetas = self.data.betaNames
        for b in myBetas:
            try:
                index = self.data.betaNames.index(b)
                values[b] = self.data.betas[index].value
            except:
                keys = ''
                for k in self.data.betaNames:
                    keys += f' {k}'
                err = f"The value of {b} is not available in the results. The following parameters are available: {keys}"
                raise excep.biogemeError(err)
        return values

    def getVarCovar(self):
        """ Obtain the Rao-Cramer variance covariance matrix as a Pandas data frame.

        :return: Rao-Cramer variance covariance matrix
        :rtype: pandas.DataFrame
        """
        names = [b.name for b in self.data.betas]
        vc = pd.DataFrame(index=names,columns=names)
        for i in range(len(self.data.betas)):
            for j in range(len(self.data.betas)):
                vc.at[self.data.betas[i].name, self.data.betas[j].name] = self.data.varCovar[i,j]
        return vc   

    
    def getRobustVarCovar(self):
        """ Obtain the robust variance covariance matrix as a Pandas data frame.

        :return: robust variance covariance matrix
        :rtype: pandas.DataFrame
        """
        names = [b.name for b in self.data.betas]
        vc = pd.DataFrame(index=names,columns=names)
        for i in range(len(self.data.betas)):
            for j in range(len(self.data.betas)):
                vc.at[self.data.betas[i].name, self.data.betas[j].name] = self.data.robust_varCovar[i,j]
        return vc   


    def getBootstrapVarCovar(self):
        """ Obtain the bootstrap variance covariance matrix as a Pandas data frame.

        :return: bootstrap variance covariance matrix, or None if not available
        :rtype: pandas.DataFrame
        """
        if self.data.bootstrap is None:
            return None
        
        names = [b.name for b in self.data.betas]
        vc = pd.DataFrame(index=names,columns=names)
        for i in range(len(self.data.betas)):
            for j in range(len(self.data.betas)):
                vc.at[self.data.betas[i].name, self.data.betas[j].name] = self.data.bootstrap_varCovar[i,j]
        return vc   
    
    def writeHtml(self):
        """ Write the results in an HTML file.
        """
        self.data.htmlFileName = bf.getNewFileName(self.data.modelName,"html")
        f = open(self.data.htmlFileName,"w")
        f.write(self.getHtml())
        self.logger.general(f"Results saved in file {self.data.htmlFileName}")
        f.close()

    def writeLaTeX(self):
        """ Write the results in a LaTeX file.
        """
        self.data.latexFileName = bf.getNewFileName(self.data.modelName,"tex")
        f = open(self.data.latexFileName,"w")
        f.write(self.getLaTeX())
        f.close()
        
    def _getHtmlHeader(self):
        """Prepare the header for the HTML file, containing comments and the
        version of Biogeme.

        Return:
           string containing the header.
        """
        h = ""
        h += "<html>"
        h += "<head>"
        h += "<script src='http://transp-or.epfl.ch/biogeme/sorttable.js'></script>"
        h += "<meta http-equiv='Content-Type' content='text/html; charset=utf-8' />"
        h += "<title>"+self.data.modelName+" - Report from biogeme "+bv.getVersion()+" ["+bv.versionDate+"]</title>"
        h += "<meta name='keywords' content='biogeme, discrete choice, random utility'>"
        h += "<meta name='description' content='Report from biogeme "+bv.getVersion()+" ["+bv.versionDate+"]'>"
        h += "<meta name='author' content='"+bv.author+"'>"
        h += "<style type=text/css>"
        h += ".biostyle"
        h += "	{font-size:10.0pt;"
        h += "	font-weight:400;"
        h += "	font-style:normal;"
        h += "	font-family:Courier;}"
        h += ".boundstyle"
        h += "	{font-size:10.0pt;"
        h += "	font-weight:400;"
        h += "	font-style:normal;"
        h += "	font-family:Courier;"
        h += "        color:red}"
        h += "</style>"
        h += "</head>"
        h += "<body bgcolor='#ffffff'>"
        return h

    def _getHtmlFooter(self):
        """Prepare the footer for the HTML file.

        Return:
           string containing the footer.
        """
        return "</html>"
        
    def getBetasForSensitivityAnalysis(self,myBetas,size=100,useBootstrap=False):
        """Generate draws from the distribution of the estimates, for
        sensitivity analysis.

        :param myBetas: names of the parameters for which draws are requested.
        :type myBetas: list(string)
        :param size: number of draws. Default: 100.
        :type size: int
        :param useBootstrap: if True, the variance-covariance matrix
                  generated by the bootstrapping is used for simulation. If
                  False, the robust variance-covariance matrix is used. Default: False.
        :type useBootstrap: bool

        :raise biogeme.exceptions.biogemeError: if useBootstrap is True and the bootstrap matrix is not available.

        :return: numpy table with as many rows as draws, and as many
           columns as parameters.
        :rtype: numpy.array

        """
        if useBootstrap and self.data.bootstrap is None:
            err = f"Bootstrap variance-covariance matrix not available for simulation. Use useBootstrap=False."
            raise excep.biogemeError(err)
            

        theMatrix = self.data.bootstrap_varCovar if useBootstrap else self.data.robust_varCovar
        simulatedBetas = np.random.multivariate_normal(self.data.betaValues, theMatrix, size, check_valid='warn')
        index = [self.data.betaNames.index(b) for b in myBetas]
        return simulatedBetas[:,index]



"""File 01oneLatentRegression.py

Measurement equation where the indicators are assumed continuous.
Linear regression.

:author: Michel Bierlaire, EPFL
:date: Mon Sep  9 16:30:04 2019

"""
import pandas as pd
import numpy as np
import biogeme.database as db
import biogeme.biogeme as bio
from biogeme.models import piecewise
import biogeme.loglikelihood as ll
import biogeme.optimization as opt
from biogeme.expressions import Beta, DefineVariable, log, Elem, bioMultSum

# Read the data
df = pd.read_csv("optima.dat",sep='\t')
database = db.Database("optima",df)

# The following statement allows you to use the names of the variable
# as Python variable.
globals().update(database.variables)

# Exclude observations such that the chosen alternative is -1
exclude = (Choice == -1)
database.remove(exclude)

# Piecewise linear definition of income
ScaledIncome = DefineVariable('ScaledIncome',\
                              CalculatedIncome / 1000,database)

thresholds = [4,6,8,10]
ContIncome = piecewise(ScaledIncome,thresholds)
ContIncome_0_4000 = ContIncome[0]
ContIncome_4000_6000 = ContIncome[1]
ContIncome_6000_8000 = ContIncome[2]
ContIncome_8000_10000 = ContIncome[3]
ContIncome_10000_more = ContIncome[4]

# Definition of other variables
age_65_more = DefineVariable('age_65_more',age >= Numeric(65),database)
moreThanOneCar = DefineVariable('moreThanOneCar',NbCar > 1,database)
moreThanOneBike = DefineVariable('moreThanOneBike',NbBicy > 1,database)
individualHouse = DefineVariable('individualHouse',\
                                 HouseType == 1,database)
male = DefineVariable('male',Gender == 1,database)
haveChildren = DefineVariable('haveChildren',\
                              ((FamilSitu == 3)+(FamilSitu == 4)) > 0,database)
haveGA = DefineVariable('haveGA',GenAbST == 1,database)
highEducation = DefineVariable('highEducation', Education >= 6,database)

# Parameters to be estimated
coef_intercept = Beta('coef_intercept',0.0,None,None,0)
coef_age_65_more = Beta('coef_age_65_more',0.0,None,None,0)
coef_age_unknown = Beta('coef_age_unknown',0.0,None,None,0)
coef_haveGA = Beta('coef_haveGA',0.0,None,None,0)
coef_ContIncome_0_4000 = \
 Beta('coef_ContIncome_0_4000',0.0,None,None,0)
coef_ContIncome_4000_6000 = \
 Beta('coef_ContIncome_4000_6000',0.0,None,None,0)
coef_ContIncome_6000_8000 = \
 Beta('coef_ContIncome_6000_8000',0.0,None,None,0)
coef_ContIncome_8000_10000 = \
 Beta('coef_ContIncome_8000_10000',0.0,None,None,0)
coef_ContIncome_10000_more = \
 Beta('coef_ContIncome_10000_more',0.0,None,None,0)
coef_moreThanOneCar = \
 Beta('coef_moreThanOneCar',0.0,None,None,0)
coef_moreThanOneBike = \
 Beta('coef_moreThanOneBike',0.0,None,None,0)
coef_individualHouse = \
 Beta('coef_individualHouse',0.0,None,None,0)
coef_male = Beta('coef_male',0.0,None,None,0)
coef_haveChildren = Beta('coef_haveChildren',0.0,None,None,0)
coef_highEducation = Beta('coef_highEducation',0.0,None,None,0)

### Latent variable: structural equation

# Note that the expression must be on a single line. In order to 
# write it across several lines, each line must terminate with 
# the \ symbol

CARLOVERS = \
            coef_intercept +\
            coef_age_65_more * age_65_more +\
            coef_ContIncome_0_4000 * ContIncome_0_4000 +\
            coef_ContIncome_4000_6000 * ContIncome_4000_6000 +\
            coef_ContIncome_6000_8000 * ContIncome_6000_8000 +\
            coef_ContIncome_8000_10000 * ContIncome_8000_10000 +\
            coef_ContIncome_10000_more * ContIncome_10000_more +\
            coef_moreThanOneCar * moreThanOneCar +\
            coef_moreThanOneBike * moreThanOneBike +\
            coef_individualHouse * individualHouse +\
            coef_male * male +\
            coef_haveChildren * haveChildren +\
            coef_haveGA * haveGA +\
            coef_highEducation * highEducation

sigma_s = Beta('sigma_s',1,None,None,1)

### Measurement equations
INTER_Envir01 = Beta('INTER_Envir01',0,None,None,1)
INTER_Envir02 = Beta('INTER_Envir02',0,None,None,0)
INTER_Envir03 = Beta('INTER_Envir03',0,None,None,0)
INTER_Mobil11 = Beta('INTER_Mobil11',0,None,None,0)
INTER_Mobil14 = Beta('INTER_Mobil14',0,None,None,0)
INTER_Mobil16 = Beta('INTER_Mobil16',0,None,None,0)
INTER_Mobil17 = Beta('INTER_Mobil17',0,None,None,0)

B_Envir01_F1 = Beta('B_Envir01_F1',-1,None,None,1)
B_Envir02_F1 = Beta('B_Envir02_F1',-1,None,None,0)
B_Envir03_F1 = Beta('B_Envir03_F1',1,None,None,0)
B_Mobil11_F1 = Beta('B_Mobil11_F1',1,None,None,0)
B_Mobil14_F1 = Beta('B_Mobil14_F1',1,None,None,0)
B_Mobil16_F1 = Beta('B_Mobil16_F1',1,None,None,0)
B_Mobil17_F1 = Beta('B_Mobil17_F1',1,None,None,0)

MODEL_Envir01 = INTER_Envir01 + B_Envir01_F1 * CARLOVERS
MODEL_Envir02 = INTER_Envir02 + B_Envir02_F1 * CARLOVERS
MODEL_Envir03 = INTER_Envir03 + B_Envir03_F1 * CARLOVERS
MODEL_Mobil11 = INTER_Mobil11 + B_Mobil11_F1 * CARLOVERS
MODEL_Mobil14 = INTER_Mobil14 + B_Mobil14_F1 * CARLOVERS
MODEL_Mobil16 = INTER_Mobil16 + B_Mobil16_F1 * CARLOVERS
MODEL_Mobil17 = INTER_Mobil17 + B_Mobil17_F1 * CARLOVERS

# We force the SIGMA to be positive by taking the exp.
SIGMA_STAR_Envir01 = exp(Beta('SIGMA_STAR_Envir01',1,None,None,0))
SIGMA_STAR_Envir02 = exp(Beta('SIGMA_STAR_Envir02',1,None,None,0))
SIGMA_STAR_Envir03 = exp(Beta('SIGMA_STAR_Envir03',1,None,None,0))
SIGMA_STAR_Mobil11 = exp(Beta('SIGMA_STAR_Mobil11',1,None,None,0))
SIGMA_STAR_Mobil14 = exp(Beta('SIGMA_STAR_Mobil14',1,None,None,0))
SIGMA_STAR_Mobil16 = exp(Beta('SIGMA_STAR_Mobil16',1,None,None,0))
SIGMA_STAR_Mobil17 = exp(Beta('SIGMA_STAR_Mobil17',1,None,None,0))

# We build a dict with each contribution to the loglikelihood if (var
# > 0) and (var < 6). If not, 0 is returned.
F = {}
F['Envir01'] = Elem({0:0, \
 1:ll.loglikelihoodregression(Envir01,MODEL_Envir01,SIGMA_STAR_Envir01)},\
  (Envir01 > 0)*(Envir01 < 6))
F['Envir02'] = Elem({0:0, \
 1:ll.loglikelihoodregression(Envir02,MODEL_Envir02,SIGMA_STAR_Envir02)},\
  (Envir02 > 0)*(Envir02 < 6))
F['Envir03'] = Elem({0:0, \
 1:ll.loglikelihoodregression(Envir03,MODEL_Envir03,SIGMA_STAR_Envir03)},\
  (Envir03 > 0)*(Envir03 < 6))
F['Mobil11'] = Elem({0:0, \
 1:ll.loglikelihoodregression(Mobil11,MODEL_Mobil11,SIGMA_STAR_Mobil11)},\
  (Mobil11 > 0)*(Mobil11 < 6))
F['Mobil14'] = Elem({0:0, \
 1:ll.loglikelihoodregression(Mobil14,MODEL_Mobil14,SIGMA_STAR_Mobil14)},\
  (Mobil14 > 0)*(Mobil14 < 6))
F['Mobil16'] = Elem({0:0, \
 1:ll.loglikelihoodregression(Mobil16,MODEL_Mobil16,SIGMA_STAR_Mobil16)},\
  (Mobil16 > 0)*(Mobil16 < 6))
F['Mobil17'] = Elem({0:0, \
 1:ll.loglikelihoodregression(Mobil17,MODEL_Mobil17,SIGMA_STAR_Mobil17)},\
  (Mobil17 > 0)*(Mobil17 < 6))

# The log likelihood is the sum of the elements of the above dict
loglike = bioMultSum(F)

# Define level of verbosity
import biogeme.messaging as msg
logger = msg.bioMessage()
#logger.setSilent()
#logger.setWarning()
logger.setGeneral()
#logger.setDetailed()

# Create the Biogeme object
biogeme  = bio.BIOGEME(database,loglike)
biogeme.modelName = "01oneLatentRegression"

# Estimate the parameters
results = biogeme.estimate(algorithm=opt.newtonTrustRegionForBiogeme)

print(f"Estimated betas: {len(results.data.betaValues)}")
print(f"final log likelihood: {results.data.logLike:.3f}")
print(f"Output file: {results.data.htmlFileName}")
results.writeLaTeX()
print(f"LaTeX file: {results.data.latexFileName}")





""" Functions to calculate the log likelihood

:author: Michel Bierlaire
:date: Fri Mar 29 17:11:44 2019

"""

from biogeme.expressions import *

def loglikelihood(prob):
    """
    Simply computes the log of the probability

    :param prob: An expression providing the value of the probability.
    :type prob: biogeme.expressions.Expression

    :return: the logarithm of the probability.
    :rtype: biogeme.expressions.Expression

    """    
    loglikelihood = log(prob)
    return loglikelihood

def mixedloglikelihood(prob):
    """Compute a simulated loglikelihood function

    :param prob: An expression providing the value of the
                 probability. Although it is not formally necessary,
                 the expression should contain one or more random
                 variables of a given distribution, and therefore
                 is defined as

    .. math:: P(i|\\xi_1,\\ldots,\\xi_L)

    
    :type prob: biogeme.expressions.Expression
    
    :return: the simulated loglikelihood, given by 

    .. math:: \\ln\\left(\\sum_{r=1}^R P(i|\\xi^r_1,\\ldots,\\xi^r_L) \\right)

    where :math:`R` is the number of draws, and :math:`\\xi_j^r` is the rth draw of the random variable :math:`\\xi_j`.

    :rtype: biogeme.expressions.Expression

    """
    l = MonteCarlo(prob)
    return log(l)

def likelihoodregression(meas,model,sigma):
    """ Computes likelihood function of a regression model.

    :param meas: An expression providing the value :math:`y` of the measure for the current observation.
    :type meas: biogeme.expressions.Expression
    :param model: An expression providing the output :math:`m` of the model for the current observation.
    :type model: biogeme.expressions.Expression
    :param sigma: An expression (typically, a parameter) providing the standard error :math:`\sigma` of the error term.
    :type sigma: biogeme.expressions.Expression
    :return: The likelihood of the regression, assuming a normal distribution, that is

    .. math:: \\frac{1}{\\sigma} \\phi\\left( \\frac{y-m}{\\sigma} \\right) 

    where :math:`\phi(\cdot)` is the pdf of the normal distribution.

    :rtype: biogeme.expressions.Expression
    """
    t = (meas - model) / sigma
    f = bioNormalPdf(t) / sigma
    return f


def loglikelihoodregression(meas,model,sigma):
    """Computes log likelihood function of a regression model.

    :param meas: An expression providing the value :math:`y` of the
                 measure for the current observation.
    :type meas: biogeme.expressions.Expression

    :param model: An expression providing the output :math:`m` of the
                  model for the current observation.
    :type model: biogeme.expressions.Expression

    :param sigma: An expression (typically, a parameter) providing the standard error :math:`\sigma` of the error term.
    :type sigma: biogeme.expressions.Expression

    :return: the likelihood of the regression, assuming a normal distribution, that is
     
    .. math:: -\\left( \\frac{(y-m)^2}{2\\sigma^2} \\right) - \\log(\\sigma) - \\frac{1}{2}\\log(2\\pi)

    :rtype: biogeme.expressions.Expression
    """
    t = (meas - model) / sigma
    f = - (t ** 2) / 2 - log(sigma) -0.9189385332
    return f




""" Implementation of the class Database, wrapping a pandas dataframe for specific services to Biogeme

:author: Michel Bierlaire

:date: Tue Mar 26 16:42:54 2019
"""

import os
from biogeme.expressions import Variable, isNumeric
import biogeme.exceptions as excep
from datetime import datetime
import numpy as np
import pandas as pd

import biogeme.filenames as bf
import biogeme.draws as draws
import biogeme.messaging as msg
import biogeme.tools as tools

class Database:
    """ Class that contains and prepare the database.
    """
    def __init__(self,name,pandasDatabase):
        """Constructor
        
        :param name: name of the database.
        :type name: string 

        :param pandasDatabase: data stored in a pandas data frame.
        :type pandasDatabase: pandas.DataFrame

        """
        self.logger = msg.bioMessage()
        start_time = datetime.now()
        ## Name of the database. Used mainly for the file name when dumping data.
        self.name = name
        ## Pandas data frame containing the data.
        self.data = pandasDatabase
        self.fullData = pandasDatabase
        # self.variables is initialized by _generateHeaders()
        self.variables = None
        self._generateHeaders()
        ## Number of observations removed by the function Database.remove
        self.excludedData = 0
        ## Name of the column identifying the individuals in a panel data context. None if data is not panel.
        self.panelColumn = None
        ## map identifying the range of observations for each individual in a panel data context. None if data is not panel.
        self.individualMap = None
        self.fullIndividualMap = None
        ## Initialize the dictionary containing random number generators with a series of native generators.
        self._initNativeRandomNumberGenerators()
        ## Dictionary containing user defined random number generators. Defined by the function Database.setRandomNumberGenerators that checks that reserved keywords are not used. The element of the dictionary is a tuple with two elements: (0) the function generating the draws, and (1) a string describing the type of draws
        self.userRandomNumberGenerators = dict()
        ## Number of draws generated by the function Database.generateDraws. Value 0 if this fumction is not called.
        self.numberOfDraws = 0
        ## Types of draws for Monte Carlo integration
        self.typesOfDraws = {}

        self._auditDone = False
        listOfErrors,listOfWarnings = self._audit()
        if listOfWarnings:
            self.logger.warning('\n'.join(listOfWarnings))
        if listOfErrors:
            self.logger.warning('\n'.join(listOfErrors))
            raise excep.biogemeError("\n".join(listOfErrors))

    def _initNativeRandomNumberGenerators(self):
        def uniform_antithetic(sampleSize,numberOfDraws):
            return draws.getAntithetic(draws.getUniform,
                                       sampleSize,
                                       numberOfDraws)
        
        def halton2(sampleSize,numberOfDraws):
            return draws.getHaltonDraws(sampleSize,
                                        numberOfDraws,
                                        base=2,
                                        skip=10)

        def halton3(sampleSize,numberOfDraws):
            return draws.getHaltonDraws(sampleSize,
                                        numberOfDraws,
                                        base=3,
                                        skip=10)

        def halton5(sampleSize,numberOfDraws):
            return draws.getHaltonDraws(sampleSize,
                                        numberOfDraws,
                                        base=5,
                                        skip=10)

        def MLHS_anti(sampleSize,numberOfDraws):
            return draws.getAntithetic(draws.getLatinHypercubeDraws,
                                       sampleSize,
                                       numberOfDraws)

        def symm_uniform(sampleSize,numberOfDraws):
            return draws.getUniform(sampleSize,numberOfDraws,symmetric=True)

        def symm_uniform_antithetic(sampleSize,numberOfDraws):
            R = int(numberOfDraws / 2)
            draws = symm_uniform(sampleSize,R)
            return np.concatenate((draws,-draws),axis=1)
        
        def symm_halton2(sampleSize,numberOfDraws):
            return draws.getHaltonDraws(sampleSize,
                                        numberOfDraws,
                                        symmetric=True,
                                        base=2,
                                        skip=10)

        def symm_halton3(sampleSize,numberOfDraws):
            return draws.getHaltonDraws(sampleSize,
                                        numberOfDraws,
                                        symmetric=True,
                                        base=3,
                                        skip=10)

        def symm_halton5(sampleSize,numberOfDraws):
            return draws.getHaltonDraws(sampleSize,
                                        numberOfDraws,
                                        symmetric=True,
                                        base=5,
                                        skip=10)

        def symm_MLHS(sampleSize,numberOfDraws):
            return draws.getLatinHypercubeDraws(sampleSize,
                                                numberOfDraws,
                                                symmetric=True)
                                       
        def symm_MLHS_anti(sampleSize,numberOfDraws):
            R = int(numberOfDraws / 2)
            draws = symm_MLHS(sampleSize,R)
            return np.concatenate((draws,-draws),axis=1)

        def normal_antithetic(sampleSize,numberOfDraws):
            return draws.getNormalWichuraDraws(sampleSize=sampleSize,
                                               numberOfDraws=numberOfDraws,
                                               antithetic=True)

        def normal_halton2(sampleSize,numberOfDraws):
            unif = draws.getHaltonDraws(sampleSize,
                                        numberOfDraws,
                                        base=2,
                                        skip=10)
            return draws.getNormalWichuraDraws(sampleSize,
                                               numberOfDraws,
                                               uniformNumbers=unif,
                                               antithetic=False)

        def normal_halton3(sampleSize,numberOfDraws):
            unif = draws.getHaltonDraws(sampleSize,
                                        numberOfDraws,
                                        base=2,
                                        skip=10)
            return draws.getNormalWichuraDraws(sampleSize,
                                               numberOfDraws,
                                               uniformNumbers=unif,
                                               antithetic=False)

        def normal_halton5(sampleSize,numberOfDraws):
            unif = draws.getHaltonDraws(sampleSize,
                                        numberOfDraws,
                                        base=2,
                                        skip=10)
            return draws.getNormalWichuraDraws(sampleSize,
                                               numberOfDraws,
                                               uniformNumbers=unif,
                                               antithetic=False)

        def normal_MLHS(sampleSize,numberOfDraws):
            unif = draws.getLatinHypercubeDraws(sampleSize,
                                                numberOfDraws)
            return draws.getNormalWichuraDraws(sampleSize,
                                               numberOfDraws,
                                               uniformNumbers=unif,
                                               antithetic=False)
        
        def normal_MLHS_anti(sampleSize,numberOfDraws):
            unif = draws.getLatinHypercubeDraws(sampleSize,
                                                int(numberOfDraws/2))
            return draws.getNormalWichuraDraws(sampleSize,
                                               numberOfDraws,
                                               uniformNumbers=unif,
                                               antithetic=True)
        
        ## Dictionary containing native random number generators.
        self.nativeRandomNumberGenerators = {
            'UNIFORM': (draws.getUniform, 'Uniform U[0,1]'),
            'UNIFORM_ANTI':(uniform_antithetic, 'Antithetic uniform U[0,1]'),
            'UNIFORM_HALTON2':(halton2, 'Halton draws with base 2, skipping the first 10'),
            'UNIFORM_HALTON3':(halton3,'Halton draws with base 3, skipping the first 10'),
            'UNIFORM_HALTON5':(halton5,'Halton draws with base 5, skipping the first 10'),
            'UNIFORM_MLHS':(draws.getLatinHypercubeDraws,'Modified Latin Hypercube Sampling on [0,1]'),
            'UNIFORM_MLHS_ANTI':(MLHS_anti,'Antithetic Modified Latin Hypercube Sampling on [0,1]'),
            'UNIFORMSYM': (symm_uniform, 'Uniform U[-1,1]'),
            'UNIFORMSYM_ANTI':(symm_uniform_antithetic, 'Antithetic uniform U[-1,1]'),
            'UNIFORMSYM_HALTON2':(symm_halton2, 'Halton draws on [-1,1] with base 2, skipping the first 10'),
            'UNIFORMSYM_HALTON3':(symm_halton3,'Halton draws on [-1,1] with base 3, skipping the first 10'),
            'UNIFORMSYM_HALTON5':(symm_halton5,'Halton draws on [-1,1] with base 5, skipping the first 10'),
            'UNIFORMSYM_MLHS':(symm_MLHS,'Modified Latin Hypercube Sampling on [-1,1]'),
            'UNIFORMSYM_MLHS_ANTI':(symm_MLHS_anti,'Antithetic Modified Latin Hypercube Sampling on [-1,1]'),
            'NORMAL': (draws.getNormalWichuraDraws,'Normal N(0,1) draws'),
            'NORMAL_ANTI':(normal_antithetic,'Antithetic normal draws'),
            'NORMAL_HALTON2':(normal_halton2,'Normal draws from Halton base 2 sequence'),
            'NORMAL_HALTON3':(normal_halton3,'Normal draws from Halton base 3 sequence'),
            'NORMAL_HALTON5':(normal_halton5,'Normal draws from Halton base 5 sequence'),
            'NORMAL_MLHS':(normal_MLHS,'Normal draws from Modified Latin Hypercube Sampling'),
            'NORMAL_MLHS_ANTI':(normal_MLHS_anti,'Antithetic normal draws from Modified Latin Hypercube Sampling')
            }

    def descriptionOfNativeDraws(self):
        """ Describe the draws available draws with Biogeme
        
        :return: dict, where the keys are the names of the draws, and the value their description

        Example of output::

        ['UNIFORM: Uniform U[0,1]','UNIFORM_ANTI: Antithetic uniform U[0,1]'],'NORMAL: Normal N(0,1) draws']         

        :rtype: dict
        """
        return [f"{key}: {tuple[1]}" for key,tuple in self.nativeRandomNumberGenerators.items()]   
        
    def _audit(self):
        """ Performs a series of checks and reports warnings and errors.
              - Check if there are non numerical entries.
              - Check if there are NaN (not a number) entries.
              - Check if there are strings.
              - Check if the numbering of individuals are contiguous (panel data only).

            Returns: 
                A tuple of two lists with the results of the diagnostic: listOfErrors, listOfWarnings
        """
        listOfErrors = []
        listOfWarnings = []
        if self._auditDone:
            return listOfErrors,listOfWarnings

        for col,dtype in self.data.dtypes.items():
            if not np.issubdtype(dtype, np.number):
                theError = "Column {} in the database does contain {}".format(col,dtype)
                listOfErrors.append(theError)
        
        if self.data.isnull().values.any():
            theError = "The database contains NaN value(s). Detect where they are using the function isnan()"
            listOfErrors.append(theError)

        self._auditDone = True
        return listOfErrors,listOfWarnings

    def _generateHeaders(self):
        """Generates the file headers.py containing the name of the headers
        of the database so that they can be used as an object of type
        biogeme.expressions.Expression
        """
        #file = open("headers.py","w")
        #print("from biogeme.expressions import Plus, Minus, Times, Divide, Power, bioMin, bioMax, And, Or, Equal, NotEqual, LessOrEqual, GreaterOrEqual, Less, Greater, UnaryMinus, MonteCarlo, bioNormalPdf, bioNormalCdf, PanelLikelihoodTrajectory, exp, log, Derive, Integrate, bioDraws, Numeric, Variable, DefineVariable, RandomVariable, Beta, bioLogLogit, bioLogLogitFullChoiceSet, bioMultSum, Elem, bioLinearUtility",file=file)
        #for c in self.data.columns:
        #    globals()[c] = Variable(c)
        #    print("{}=Variable('{}')".format(c,c),file=file)
        #file.close()

        self.variables = {col: Variable(col) for col in self.data.columns}
        file = open("headers.py","w")
        print("import biogeme.exceptions as excep",file=file)
        print("raise excep.biogemeError('The file headers.py must not be included anymore. Replace the statement\\nfrom headers import *\\nby\\nfrom biogeme.expressions import Beta, DefineVariable\\nglobals().update(database.variables)\\n\\nReplace also any occurence of bioLogLogit by biogeme.models.loglogit.')",file=file)
        file.close()
        #for c in self.data.columns:
        #    globals()[c] = Variable(c)
        #    print("{}=Variable('{}')".format(c,c),file=file)

    def valuesFromDatabase(self,expression):
        """ Evaluates an expression for each entry of the database.

        :param expression: expression to evaluate
        :type expression:  biogeme.expressions.Expression. 

        :return: numpy series, long as the number of entries in the database, containing the calculated quantities.
        :rtype: numpy.Series

        """
        self._expression = expression
        def functionToApply(row):
            self._expression.setRow(row)
            res = self._expression.getValue()
            return res
        res =  self.data.apply(functionToApply,axis=1)
        return res

    def checkAvailabilityOfChosenAlt(self,avail,choice):
        """ Check if the chosen alternative is available for each entry in the database.

        :param avail: list of expressions to evaluate the availability conditions for each alternative.
        :type avail: list of biogeme.expressions.Expression
        :param choice: expression for the chosen alternative.
        :type choice: biogeme.expressions.Expression 

        :return: numpy series of bool, long as the number of entries in the database, containing True is the chosen alternative is available, False otherwise.
        :rtype: numpy.Series
        """
        self._avail = avail
        self._choice = choice
        def functionToApply(row):
            self._choice.setRow(row)
            chosen = self._choice.getValue()
            try:
                avExpression = self._avail[chosen]
            except:
                return False
            avExpression.setRow(row)
            av = avExpression.getValue()
            return av != 0
        res = self.data.apply(functionToApply,axis=1)
        return res
            
    def sumFromDatabase(self,expression):
        """ Calculates the value of an expression for each entry in the database, and returns the sum.

        :param expression: expression to evaluate 
        :type expression: biogeme.expressions.Expression 
        :return: sum of the expressions over the database.
        :rtype: float
        """
        self._expression = expression
        def functionToApply(row):
            self._expression.setRow(row)
            res = self._expression.getValue()
            return res
        res =  np.nansum(self.data.apply(functionToApply,axis=1))
        return res

    def scaleColumn(self,column,scale):
        """Multiply an entire column by a scale value

        :param column: name of the column
        :type column: string
        :param scale: value of the scale. All values of the column will
              be multiplied by that scale.
        :type scale: float

        """
        self.data[column] = self.data[column] * scale 

    def suggestScaling(self):
        """Suggest a scaling of the variables in the database. 

        For each column, :math:`\\delta` is the difference between the
        largest and the smallest value, or one if the difference is
        smaller than one. The level of magnitude is evaluated as a
        power of 10. The suggested scale is the inverse of this value.

        .. math:: s = \\frac{1}{10^{|\\log_{10} \\delta|}}

        where :math:`|x|` is the integer closest to :math:`x`.
        
        :return: A Pandas dataframe where each row contains the name
                 of the variable and the suggested scale s. Ideally,
                 the column should be multiplied by s.

        :rtype: pandas.DataFrame

        """
        largestValue = [max(np.abs(self.data[col].max()),np.abs(self.data[col].min())) for col in self.data.columns]
        res = [[col,1 / 10**np.round(np.log10(max(1.0,l)))] for col,l in zip(self.data.columns,largestValue)]
        df = pd.DataFrame(res, columns = ['Column', 'Scale']) 
        # Remove entries where the suggested scale is 1
        remove = (df.Scale == 1)
        df.drop(df[remove].index,inplace=True)
        return df
        
    def sampleWithReplacement(self,size=None):
        """ Extract a random sample from the database, with replacement. 

        Useful for bootstrapping. 

        :param size: size of the sample. If None, a sample of the same size as the database will be generated. Default: None.
        :type size: int

        :return: pandas dataframe with the sample.
        :rtype: pandas.DataFrame

        """
        if size is None:
            size = len(self.data)
        sample = self.data.iloc[np.random.randint(0, len(self.data), size=size)]
        return sample

    def sampleIndividualMapWithReplacement(self,size=None):
        """ Extract a random sample of the individual map from a panel data database, with replacement.

        Useful for bootstrapping. 

        
        :param size: size of the sample. If None, a sample of the same size as the database will be generated. Default: None.
        :type size: int

        :return: pandas dataframe with the sample.
        :rtype: pandas.DataFrame

        """
        if not self.isPanel():
            raise excep.biogemeError("Function sampleIndividualMapWithReplacement is available only on panel data.")
            
        if size is None:
            size = len(self.individualMap)
        sample = self.individualMap.iloc[np.random.randint(0, len(self.individualMap), size=size)]
        return sample


    def sampleWithoutReplacement(self,samplingRate,columnWithSamplingWeights=None):
        """ Replace the data set by a sample for stochastic algorithms
        
        :param samplingRate: the proportion of data to include in the sample.
        :type samplingRate: float
        :param columnWithSamplingWeights: name of the column with the sampling weights. If None, each row has eqaul probability.
        :param columnWithSamplingWeights: string

        :return: None
        """
        if self.isPanel():
            if self.fullIndividualMap is None:
                self.fullIndividualMap = self.individualMap
            else:
                # Check if the structure has not been modified since last sample
                if set(self.fullIndividualMap.columns) != set(self.individualMap.columns):
                    message = "The structure of the database has been modified since last sample. "
                    left = set(self.fullIndividualMap.columns).difference(set(self.individualMap.columns))
                    if left:
                        message += f" Columns that disappeared: {left}"
                    right = set(self.individualMap.columns).difference(set(self.fullIndividualMap.columns))
                    if right:
                        message += f" Columns that were added: {right}"
                    raise excep.biogemeError(message)

            self.individualMap = self.fullIndividualMap.sample(frac=samplingRate,weights=columnWithSamplingWeights)
            self.logger.debug(f'Full data: {self.fullIndividualMap.shape} Sampled data: {self.individualMap.shape}')
            
        else:
            # Cross sectional data
            if self.fullData is None:
                self.fullData = self.data
            else:
                # Check if the structure has not been modified since last sample
                if set(self.fullData.columns) != set(self.data.columns):
                    message = "The structure of the database has been modified since last sample. "
                    left = set(self.fullData.columns).difference(set(self.data.columns))
                    if left:
                        message += f" Columns that disappeared: {left}"
                    right = set(self.data.columns).difference(set(self.fullData.columns))
                    if right:
                        message += f" Columns that were added: {right}"
                    raise excep.biogemeError(message)

            self.data = self.fullData.sample(frac=samplingRate,weights=columnWithSamplingWeights)
            self.logger.debug(f'Full data: {self.fullData.shape} Sampled data: {self.data.shape}')

    def useFullSample(self):
        """ Re-establish the full sample for calculation of the likelihood
        """
        if self.isPanel():
            if self.fullIndividualMap is None:
                raise excep.biogemeError("Full panel data set has not been saved.")
            self.individualMap = self.fullIndividualMap
        else:
            if self.fullData is None:
                raise excep.biogemeError("Full data set has not been saved.")
            self.data = self.fullData
        
    def addColumn(self,expression,column):
        """ Add a new column in the database, calculated from an expression.

        :param expression:  expression to evaluate
        :type expression: biogeme.expressions.Expression

        :param column: name of the column to add
        :type column: string
        
        :return: the added column
        :rtype: numpy.Series

        :raises ValueError: if the column name already exists.

        """
        if column in self.data.columns:
            raise ValueError(f"Column {column} already exists in the database {self.name}")

        def functionToApply(row):
            self._expression.setRow(row)
            return self._expression.getValue()
        self._expression = expression
        self.data[column] = self.data.apply(functionToApply,axis=1)
        return self.data[column]

    def remove(self,expression):
        """ Removes from the database all entries such that the value of the expression is not 0. 
        
        :param expression: expression to evaluate
        :type expression: biogeme.expressions.Expression

        """
        columnName = '__bioRemove__'
        if isNumeric(expression):
            self.addColumn(Numeric(expression),columnName)
        else:
            self.addColumn(expression,columnName)
        self.excludedData = len(self.data[self.data[columnName] != 0].index)
        self.data.drop(self.data[self.data[columnName] != 0].index, inplace=True)
        self.data.drop(columns=[columnName],inplace=True)

    def dumpOnFile(self):
        """ Dumps the database in a CSV formatted file.

        :return:  name of the file
        :rtype: string
        """
        theName = "{}_dumped".format(self.name)
        dataFileName = bf.getNewFileName(theName,"dat")
        self.data.to_csv(dataFileName,sep='\t',index_label="__rowId")
        self.logger.general("File {} has been created".format(dataFileName))
        return dataFileName

    def setRandomNumberGenerators(self,rng):
        """Defines user-defined random numbers generators.
        
        :param rng: a dictionary of generators. The keys of the dictionary
           characterize the name of the generators, and must be
           different from the pre-defined generators in Biogeme:
           NORMAL, UNIFORM and UNIFORMSYM. The elements of the
           dictionary are functions that take two arguments: the
           number of series to generate (typically, the size of the
           database), and the number of draws per series.
        :type rng: dict

        Example::
        
            def logNormalDraws(sampleSize,numberOfDraws):
                return np.exp(np.random.randn(sampleSize,numberOfDraws))

            def exponentialDraws(sampleSize,numberOfDraws):
                return -1.0 * np.log(np.random.rand(sampleSize,numberOfDraws))

            # We associate these functions with a name
            dict = {'LOGNORMAL':(logNormalDraws,'Draws from lognormal distribution'),'EXP':(exponentialDraws,'Draws from exponential distributions')}
            myData.setRandomNumberGenerators(dict)
        
 
        """
        for k in self.nativeRandomNumberGenerators:
            if k in rng:
                raise ValueError(f"{k} is a reserved keyword for draws and cannot be used for user-defined generators") ;

        self.userRandomNumberGenerators = rng ;
    
    def generateDraws(self,types,names,numberOfDraws):
        """Generate draws for each variable.
        

        :param types: A dict indexed by the names of the variables,
                      describing the types of draws. Each of them can
                      be a native type or any type defined by the
                      function database.setRandomNumberGenerators
        :type types: dict

        :param names: the list of names of the variables that require draws to be generated.
        :type names: list of strings

        :param numberOfDraws: number of draws to generate.
        :type numberOfDraws: int

        :return: a 3-dimensional table with draws. The 3 dimensions are

              1. number of individuals
              2. number of draws
              3. number of variables

        :rtype: numpy.array

        Example::

              types = {'randomDraws1': 'NORMAL_MLHS_ANTI', 'randomDraws2': 'UNIFORM_MLHS_ANTI', 'randomDraws3': 'UNIFORMSYM_MLHS_ANTI'}
              theDrawsTable = myData.generateDraws(types,['randomDraws1','randomDraws2','randomDraws3'],10)        

        """

        self.numberOfDraws = numberOfDraws
        # Dimensions of the draw table:
        # 1. number of variables
        # 2. number of individuals
        # 3. number of draws
        listOfDraws = [None]*len(names)
        for i in range(len(names)):
            name = names[i]
            type = types[name]
            self.typesOfDraws[name] = type
            theGenerator = self.nativeRandomNumberGenerators.get(type)
            if theGenerator is None:
                theGenerator = self.userRandomNumberGenerators.get(type)
                if theGenerator is None:
                    native = [k for k in self.nativeRandomNumberGenerators]
                    user = [k for k in self.userRandomNumberGenerators]
                    raise excep.biogemeError(f"Unknown type of draws for variable {name}: {type}. Native types: {native}. User defined: {user}")
            listOfDraws[i] = theGenerator[0](self.getSampleSize(),numberOfDraws)
            if listOfDraws[i].shape != (self.getSampleSize(),numberOfDraws):
                raise excep.biogemeError(f"The draw generator for {name} must generate a numpy array of dimensions ({self.getSampleSize()},{numberOfDraws}) instead of {listOfDraws[i].shape}")

        self.theDraws = np.array(listOfDraws)
        ## Draws as a three-dimensional numpy series. The dimensions are organized to be more
        # suited for calculation. 
        # 1. number of individuals
        # 2. number of draws
        # 3. number of variables
        self.theDraws = np.moveaxis(self.theDraws,0,-1)
        return self.theDraws

    def getNumberOfObservations(self):
        """ 
          Reports the number of observations in the database. 

        Note that it returns the same value, irrespectively if the database contains panel data or not.    
        
        :return: Number of observations.
        :rtype: int

        See also:  getSampleSize()
        """
        return self.data.shape[0]
            
    def getSampleSize(self):
        """Reports the size of the sample. 

        If the data is cross-sectional, it is the number of
        observations in the database. If the data is panel, it is the
        number of individuals.

        :return: Sample size.
        :rtype: int

        See also: getNumberOfObservations()

        """
        if self.isPanel():
            return self.individualMap.shape[0]
        else:
            return self.data.shape[0]


    def isPanel(self):
        """ Tells if the data is panel or not.

        :return: True if the data is panel. 
        :rtype: bool
        """
        return self.panelColumn is not None
    
    def panel(self,columnName):
        """ Defines the data as panel data

        :param columnName: name of the columns that identifies individuals.
        :type columnName: string

        """

        self.panelColumn = columnName
        
        # Check if the data is organized in consecutive entries 
        # Number of groups of data
        nGroups = tools.countNumberOfGroups(self.data,self.panelColumn)
        sorted = self.data.sort_values(by=[self.panelColumn])
        nIndividuals = tools.countNumberOfGroups(sorted,self.panelColumn)
        if nGroups != nIndividuals:
            theError = f"The data must be sorted so that the data for the same individual are consecutive. There are {nIndividuals} individuals in the sample, and {nGroups} groups of data for column {self.panelColumn}."
            raise excep.biogemeError(theError)

        
        self._buildPanelMap()

    def _buildPanelMap(self):
        """Sorts the data so that the observations for each individuals are
contiguous, and builds a map that identifies the range of indices of
the observations of each individuals.
        """
        if self.panelColumn is not None:
            self.data = self.data.sort_values(by=self.panelColumn)
            # It is necessary to renumber the row to reflect the new ordering
            self.data.index = range(len(self.data.index))        
            map = dict()
            individuals = self.data[self.panelColumn].unique()
            for i in individuals:
                indices = self.data.loc[self.data[self.panelColumn] == i].index
                map[i] = [min(indices), max(indices)]
            self.individualMap = pd.DataFrame(map).T
            self.fullIndividualMap = self.individualMap

    def count(self,columnName,value):
        """Counts the number of observations that have a specific value in a
        given column.

        :param columnName: name of the column.
        :type columnName: string
        :param value: value that is seeked.
        :type value: float

        :return: Number of times that the value appears in the column.
        :rtype: int
        """
        return self.data[self.data[columnName] == value].count()[columnName]

    def __str__(self):
        """ Allows to print the dabase
        """
        result = f'biogeme database {self.name}:\n{self.data}'
        if self.isPanel():
            result += f'\nPanel data\n{self.individualMap}'
        return result



import biogeme as bio
from unidecode import unidecode
import datetime

### Release date
versionDate = f"{datetime.date.today()}"
### Major version number
versionMajor = 3
### Minor version number
versionMinor = 2
### Release name
versionRelease = '5'
author = "Michel Bierlaire"
urlAuthor = "http://people.epfl.ch/michel.bierlaire"
department = "Transport and Mobility Laboratory"
urlDepartment = "http://transp-or.epfl.ch"
university = "Ecole Polytechnique Fédérale de Lausanne (EPFL)"
urlUniversity = "http://www.epfl.ch"
urlBiogeme = "http://biogeme.epfl.ch"
urlForum = "https://groups.google.com/d/forum/biogeme"

def getVersion():
    """
     Version of the software

    :return:  version number, and the release.
    :rtype: string
    """
    v = f'{versionMajor}.{versionMinor}.{versionRelease}'
    return v

__version__ = getVersion()

def getHtml():
    """ Package information in HTML format
    
    :return: HTML code.
    :rtype: string

    """
    h = "<p>biogeme "+getVersion()+" ["+versionDate+"]</p>\n"
    h += "<p><a href='https://www.python.org/' target='_blank'>Python</a> package</p>\n"
    h += "<p>Home page: <a href='"+urlBiogeme+"' target='_blank'>"+urlBiogeme+"</a></p>\n"
    h += "<p>Submit questions to <a href='"+urlForum+"' target='_blank'>"+urlForum+"</a></p>\n"
    h += "<p><a href='"+urlAuthor+"'>"
    h += author
    h += "</a>, <a href='"
    h += urlDepartment
    h += "'>"
    h += department
    h += "</a>, <a href='"
    h += urlUniversity
    h += "'>"
    h += university.encode('ascii', 'xmlcharrefreplace').decode()
    h += "</a></p>\n"
    return h

def getText():
    """ Package information in text format
    
    :return: package information
    :rtype: string
    """
    
    h = "biogeme "+getVersion()+" ["+versionDate+"]\n"
    h += "Version entirely written in Python\n"
    h += "Home page: "+urlBiogeme+"\n"
    h += "Submit questions to "+urlForum+"\n"
    h += author+", "+department+", "+university+"\n"
    return h


def getLaTeX():
    """ Package information in LaTeX format
    
    :return: LaTeX comments
    :rtype: string
    """
    h = "%% biogeme "+getVersion()+" ["+versionDate+"]\n"
    h += "%% Version entirely written in Python\n"
    h += "%% Home page: "+urlBiogeme+"\n"
    h += "%% Submit questions to "+urlForum+"\n"
    h += "%% "+author+", "+department+", "+university+"\n"
    return h




"""File 12panelIntegral.py

:author: Michel Bierlaire, EPFL
:date: Sun Sep  8 18:55:38 2019

 Example of a mixture of logit models, using numerical integration.
 The datafile is organized as panel data.
 Three alternatives: Train, Car and Swissmetro
 SP data
"""
import pandas as pd
import biogeme.database as db
import biogeme.biogeme as bio
import biogeme.models as models
import biogeme.distributions as dist
from biogeme.expressions import Beta, DefineVariable, RandomVariable, PanelLikelihoodTrajectory, Integrate, log

# Read the data
df = pd.read_csv("swissmetro.dat",sep='\t')
database = db.Database("swissmetro",df)

# They are organized as panel data. The variable ID identifies each individual.
database.panel("ID")

# The Pandas data structure is available as database.data. Use all the
# Pandas functions to invesigate the database
#print(database.data.describe())

globals().update(database.variables)

# Removing some observations can be done directly using pandas.
#remove = (((database.data.PURPOSE != 1) & (database.data.PURPOSE != 3)) | (database.data.CHOICE == 0))
#database.data.drop(database.data[remove].index,inplace=True)

# Here we use the "biogeme" way for backward compatibility
exclude = (( PURPOSE != 1 ) * (  PURPOSE   !=  3  ) +  ( CHOICE == 0 )) > 0
database.remove(exclude)

# Parameters to be estimated
ASC_CAR = Beta('ASC_CAR',0,None,None,0)
ASC_TRAIN = Beta('ASC_TRAIN',0,None,None,0)
ASC_SM = Beta('ASC_SM',0,None,None,1)
B_TIME = Beta('B_TIME',0,None,None,0)
B_COST = Beta('B_COST',0,None,None,0)
B_TIME_S = Beta('B_TIME_S',0.1,None,None,0)

# Define a random parameter, normally distirbuted, designed to be used
# for numerical simulation
omega = RandomVariable('omega')
B_TIME_RND = B_TIME + B_TIME_S * omega
density = dist.normalpdf(omega) 

# Definition of new variables
SM_COST =  SM_CO   * (  GA   ==  0  ) 
TRAIN_COST =  TRAIN_CO   * (  GA   ==  0  )

# Definition of new variables: adding columns to the database 
TRAIN_TT_SCALED = DefineVariable('TRAIN_TT_SCALED',\
                                 TRAIN_TT / 100.0,database)
TRAIN_COST_SCALED = DefineVariable('TRAIN_COST_SCALED',\
                                   TRAIN_COST / 100,database)
SM_TT_SCALED = DefineVariable('SM_TT_SCALED', SM_TT / 100.0,database)
SM_COST_SCALED = DefineVariable('SM_COST_SCALED', SM_COST / 100,database)
CAR_TT_SCALED = DefineVariable('CAR_TT_SCALED', CAR_TT / 100,database)
CAR_CO_SCALED = DefineVariable('CAR_CO_SCALED', CAR_CO / 100,database)

# Definition of the utility functions
V1 = ASC_TRAIN + \
     B_TIME_RND * TRAIN_TT_SCALED + \
     B_COST * TRAIN_COST_SCALED
V2 = ASC_SM + \
     B_TIME_RND * SM_TT_SCALED + \
     B_COST * SM_COST_SCALED
V3 = ASC_CAR + \
     B_TIME_RND * CAR_TT_SCALED + \
     B_COST * CAR_CO_SCALED

# Associate utility functions with the numbering of alternatives
V = {1: V1,
     2: V2,
     3: V3}

# Associate the availability conditions with the alternatives
CAR_AV_SP =  DefineVariable('CAR_AV_SP',CAR_AV  * (  SP   !=  0  ),database)
TRAIN_AV_SP =  DefineVariable('TRAIN_AV_SP',TRAIN_AV  * (  SP   !=  0  ),database)
av = {1: TRAIN_AV_SP,
      2: SM_AV,
      3: CAR_AV_SP}

# Conditional to omega, the likelihood of one observation is
# given by the logit model (called the kernel)
obsprob = models.logit(V,av,CHOICE)

# Conditional to omega, the likelihood of all observations for
# one individual (the trajectory) is the product of the likelihood of
# each observation.
condprobIndiv = PanelLikelihoodTrajectory(obsprob)

# We integrate over omega using numerical integration
logprob = log(Integrate(condprobIndiv * density,'omega'))

# Define level of verbosity
import biogeme.messaging as msg
logger = msg.bioMessage()
#logger.setSilent()
#logger.setWarning()
logger.setGeneral()
#logger.setDetailed()

# Create the Biogeme object
biogeme  = bio.BIOGEME(database,logprob)
biogeme.modelName = "12panelIntegral"

# The numerical integration of the Rao-Cramer variance-covariance
# matrix has numerical problems. Therefore, we rely on bootstrapping
# to calculate the statistics.
results = biogeme.estimate(bootstrap=10)
pandasResults = results.getEstimatedParameters()
print(pandasResults)






import unittest
import biogeme.database as db
import pandas as pd
import numpy as np
from biogeme.expressions import *
from pathlib import Path
import os

class testDatabase(unittest.TestCase):
    def setUp(self):
        np.random.seed(90267)
        self.df = pd.DataFrame({'Person':[1,1,1,2,2],'Exclude':[0,0,1,0,1],'Variable1':[1,2,3,4,5],'Variable2':[10,20,30,40,50],'Choice':[1,2,3,1,2],'Av1':[0,1,1,1,1],'Av2':[1,1,1,1,1],'Av3':[0,1,1,1,1]})
        self.myData = db.Database('test',self.df)
        self.myPanelData = db.Database('test',self.df)
        self.Variable1=Variable('Variable1')
        self.Variable2=Variable('Variable2')
        self.Av1=Variable('Av1')
        self.Av2=Variable('Av2')
        self.Av3=Variable('Av3')
        self.Choice=Variable('Choice')
    
    def test_valuesFromDatabase(self):
        expr = self.Variable1 + self.Variable2
        result = self.myData.valuesFromDatabase(expr).tolist()
        self.assertListEqual(result,[11,22,33,44,55])

    def test_checkAvailabilityOfChosenAlt(self):
        avail = {1:self.Av1,2:self.Av2,3:self.Av3}
        result = self.myData.checkAvailabilityOfChosenAlt(avail,self.Choice).tolist()
        self.assertListEqual(result,[False,True,True,True,True])

    def test_sumFromDatabase(self):
        expression = self.Variable2 / self.Variable1
        result = self.myData.sumFromDatabase(expression)
        self.assertEqual(result,50)

    def test_addColumn(self):
        Variable1=Variable('Variable1')
        Variable2=Variable('Variable2')
        expression = Variable2 * Variable1
        result = self.myData.addColumn(expression,'NewVariable')
        theList = result.tolist()
        self.assertListEqual(theList,[10, 40, 90, 160, 250])

    def test_count(self):
        c = self.myData.count('Person',1)
        self.assertEqual(c,3)

    def test_remove(self):
        Exclude=Variable('Exclude')
        self.myData.remove(Exclude)
        rows,col = self.myData.data.shape
        self.assertEqual(rows,3)

    def test_dumpOnFile(self):
        f = self.myData.dumpOnFile()
        exists = Path(f).is_file()
        os.remove(f)
        self.assertTrue(exists)

    def test_generateDraws(self):
        randomDraws1 = bioDraws('randomDraws1','NORMAL')
        randomDraws2 = bioDraws('randomDraws2','UNIFORMSYM')
        # We build an expression that involves the two random variables
        x = randomDraws1 + randomDraws2
        types = x.dictOfDraws()
        theDrawsTable = self.myData.generateDraws(types,['randomDraws1','randomDraws2'],10)
        dim = theDrawsTable.shape
        self.assertTupleEqual(dim, (5, 10, 2))

    def test_setRandomGenerators(self):
        def logNormalDraws(sampleSize,numberOfDraws):
            return np.exp(np.random.randn(sampleSize,numberOfDraws))

        def exponentialDraws(sampleSize,numberOfDraws):
            return -1.0 * np.log(np.random.rand(sampleSize,numberOfDraws))

        # We associate these functions with a name
        dict = {'LOGNORMAL':(logNormalDraws,'Draws from lognormal distribution'),'EXP':(exponentialDraws,'Draws from exponential distributions')}
        self.myData.setRandomNumberGenerators(dict)

        # We can now generate draws from these distributions
        randomDraws1 = bioDraws('randomDraws1','LOGNORMAL')
        randomDraws2 = bioDraws('randomDraws2','EXP')
        x = randomDraws1 + randomDraws2
        types = x.dictOfDraws()
        theDrawsTable = self.myData.generateDraws(types,['randomDraws1','randomDraws2'],10)
        dim = theDrawsTable.shape
        self.assertTupleEqual(dim, (5, 10, 2))

    def test_sampleWithReplacement(self):
        res1 = self.myData.sampleWithReplacement()
        res2 = self.myData.sampleWithReplacement(12)
        dim1 = res1.shape
        dim2 = res2.shape
        self.assertTupleEqual(dim1, (5, 8))
        self.assertTupleEqual(dim2, (12, 8))

    def test_panel(self):
        # Data is not considered panel yet
        shouldBeFalse = self.myPanelData.isPanel()
        self.myPanelData.panel('Person')
        shouldBeTrue = self.myPanelData.isPanel()
        self.assertTrue(shouldBeTrue)
        self.assertFalse(shouldBeFalse)

    def test_panelDraws(self):
        randomDraws1 = bioDraws('randomDraws1','NORMAL')
        randomDraws2 = bioDraws('randomDraws2','UNIFORMSYM')
        # We build an expression that involves the two random variables
        x = randomDraws1 + randomDraws2
        types = x.dictOfDraws()
        self.myPanelData.panel('Person')
        theDrawsTable = self.myPanelData.generateDraws(types,['randomDraws1','randomDraws2'],10)
        dim = theDrawsTable.shape
        self.assertTupleEqual(dim, (2, 10, 2))

    def test_getNumberOfObservations(self):
        self.myPanelData.panel('Person')
        self.assertEqual(self.myData.getNumberOfObservations(),5)
        self.assertEqual(self.myPanelData.getNumberOfObservations(),5)

    def test_samplesSize(self):
        self.myPanelData.panel('Person')
        self.assertEqual(self.myData.getSampleSize(),5)
        self.assertEqual(self.myPanelData.getSampleSize(),2)

    def test_sampleIndividualMapWithReplacement(self):
        self.myPanelData.panel('Person')
        res = self.myPanelData.sampleIndividualMapWithReplacement(10)
        dim = res.shape
        self.assertTupleEqual(dim, (10, 2))
        
if __name__ == '__main__':
    unittest.main()
